#include <stdarg.h>
#include <sys\socket.h>
#define INCL_DOSPROCESS
#define INCL_DOSERRORS
#define INCL_DOSSEMAPHORES
#define INCL_DOSMISC
#define INCL_WIN
#define INCL_GPI
#define INCL_DOSNLS
#include <os2.h>
#include <rfb/rfbclient.h>
#include <ctype.h>
#include <os2xkey.h>
#include <utils.h>
#include <conv.h>
#include "clntconn.h"
#include "clntwnd.h"
#include "resource.h"
#include "linkseq.h"
#include "filexfer.h"
#include <debug.h>

#define _TAG_LRCONV        ((PVOID)0x00010002)
#define _WAIT_THREAD_TIME  2500
#define _LASTLOGENTRYONLY  1

#ifndef _LASTLOGENTRYONLY
typedef struct _LOGREC {
  SEQOBJ               seqObj;
  time_t               timeRec;
  CHAR                 acText[1];
} LOGREC, *PLOGREC;
#endif

typedef struct _KEYSEQ {
  ULONG                ulCount;
  ULONG                ulIndex;
  ULONG                aulKeys[1];
} KEYSEQ, *PKEYSEQ;

typedef struct _CLNTCONN {
  SEQOBJ               seqObj;
  int                  argc;
  char                 *argv[16];
  CHAR                 acArg[256];
  int                  iRFBTid;
  HMTX                 hmtxRFB;
  ULONG                ulRFBState;
  rfbClient            *pRFBClient;
  PCCLOGONINFO         pLogonInfo;
  HWND                 hwnd;
  HPS                  hpsMem;
  HBITMAP              hbmMem;
  int                  iRFBButtons;
  HPOINTER             hptrPointer;
  PKEYSEQ              pRFBKeySeq;
#ifdef _LASTLOGENTRYONLY
  PCHAR                pcLastLogEnt;
  ULONG                cbLastLogEntMax;
#else
  LINKSEQ              lsLog;
#endif
  XKFROMMP             stXKFromMP;
  ULONG                ulStopSigTime;
  HAB                  hab;
} CLNTCONN;

typedef struct _CHARTOKEYSYM {
  USHORT               usChar;
  ULONG                ulKeysym;
} CHARTOKEYSYM, *PCHARTOKEYSYM;

typedef unsigned long long       ULLONG;


static LINKSEQ         lsCCList;
static HMTX            hmtxCCList = NULLHANDLE;
static PXKBDMAP        pKbdMap = NULL;

static VOID _RFBClientCleanUp(rfbClient *pRFBClient)
{
  if ( pRFBClient == NULL )
    return;

  debugCP( "Enter" );
  cnvLRDestroy( (PLRCONV)rfbClientGetClientData( pRFBClient, _TAG_LRCONV ) );
  rfbClientSetClientData( pRFBClient, _TAG_LRCONV, NULL );

  if ( pRFBClient->listenAddress != NULL )
    free( pRFBClient->listenAddress );

  if ( pRFBClient->frameBuffer != NULL )
    free( pRFBClient->frameBuffer );

  if ( pRFBClient->appData.encodingsString != NULL )
    free( (char *)pRFBClient->appData.encodingsString );

  fxClean( pRFBClient );
  rfbClientCleanup( pRFBClient );
}

static VOID _ccDestoryBitmap(PCLNTCONN pCC)
{
  if ( pCC->hpsMem != NULLHANDLE )
  {
    HDC      hdcMem = GpiQueryDevice( pCC->hpsMem );

    if ( pCC->hpsMem != NULLHANDLE )
    {
      if ( !GpiDestroyPS( pCC->hpsMem ) )
        debug( "GpiDestroyPS() failed" );
      pCC->hpsMem = NULLHANDLE;
    }

    if ( DevCloseDC( hdcMem ) == DEV_ERROR )
      debug( "DevCloseDC() failed" );
  }

  if ( pCC->hbmMem != NULLHANDLE )
  {
    if ( !GpiDeleteBitmap( pCC->hbmMem ) )
      debug( "GpiDeleteBitmap() failed" );
    pCC->hbmMem = NULLHANDLE;
  }
}

static VOID _conv8to16(PBYTE pbDst, PBYTE pbSrc, ULONG cPixels)
{
  register BYTE        bR, bG, bB;

  for( ; cPixels > 0; cPixels--, pbDst += 2, pbSrc++ )
  {
    bB = ((*pbSrc >> 6) & 0x03);
    bG = ((*pbSrc >> 3) & 0x07);
    bR = (*pbSrc & 0x07);

    *((PUSHORT)pbDst) = 
      ( ( bR << 13 ) | ( (bR & 6) << 10 ) |
      ( ( bB << 3 ) | ( bB << 1 ) | ( bB >> 1 ) ) |
      ( ( bG << 8 ) | ( bG << 5 ) ) );
  }
}

static VOID _conv8to32(PBYTE pbDst, PBYTE pbSrc, ULONG cPixels)
{
  register BYTE        bR, bG, bB;

  for( ; cPixels > 0; cPixels--, pbDst += 4, pbSrc++ )
  {
    bB = ((*pbSrc >> 6) & 0x03);
    bG = ((*pbSrc >> 3) & 0x07);
    bR = (*pbSrc & 0x07);

    *((PULONG)pbDst) = 
      ( ( bR << 21 ) | ( bR << 18 ) | ( (bR & 6) << 15 ) ) |
      ( ( bG << 13 ) | ( bG << 10 ) | ( (bG & 6) << 7 ) ) |
      ( ( bB << 6 ) | ( bB << 4 ) | ( bB << 2 ) | bB );
  }
}

static VOID _ccSetThreadState(PCLNTCONN pCC, ULONG ulState)
{
  ULONG  ulRC;
  debug( "New state: %u", ulState );

  ulRC = DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );
  if ( ulRC != NO_ERROR )
    debug( "DosRequestMutexSem(), rc = %lu", ulRC );

  if ( pCC->ulRFBState == RFBSTATE_FINISH )
    debugCP( "WTF?! thread state already is RFBSTATE_FINISH" );

  pCC->ulRFBState = ulState;
  if ( ulState == RFBSTATE_FINISH )
    pCC->iRFBTid = -1;

  /* We use WinPostMsg() to post a message to the window. If you use
     WinSendMsg(), the mutex may deadlock in the window's message handler
     (if, for example, ccLockReadyState() is called).
     Remember, _ccSetThreadState() can be called with mutex already locked. */
  WinPostMsg( pCC->hwnd, WM_VNC_STATE, MPFROMP(pCC), MPFROMLONG(ulState) );

  DosReleaseMutexSem( pCC->hmtxRFB );
}


// libvncclient call back routines.
// --------------------------------

/* Called from _cbRFBGetPassword() and _cbRFBGetCredential() to safety obtain
   logon information from CLNTCONN object.  */
static BOOL _logonInfoToThread(PCLNTCONN pCC, PCCLOGONINFO pLogonInfo)
{
  _ccSetThreadState( pCC, RFBSTATE_WAITUSER );

  if ( !(BOOL)WinSendMsg( pCC->hwnd, WM_VNC_QUERYLOGON,
                          MPFROMP(pCC), MPFROMLONG(pLogonInfo) ) )
  {
    debug( "Logon information has not been set" );
    ccDisconnectSignal( pCC );
    return FALSE;
  }

  debug( "Back to state RFBSTATE_INITPROGRESS" );
  _ccSetThreadState( pCC, RFBSTATE_INITPROGRESS );

  return TRUE;
}

static rfbBool _cbRFBResize(rfbClient *pRFBClient)
{
  PCLNTCONN  pCC = (PCLNTCONN)rfbClientGetClientData( pRFBClient, NULL );
  size_t     sizeAlloc;
  ULONG      ulBmpBPP;

  /* 2023-12-21 When working with UltraVNC server and scaling, this callback
     function is called twice - with the framebuffer size, then with the
     unscaled desktop size. Avoid increasing the frame buffer size.  */
// -1-
  if ( pCC->hbmMem != NULLHANDLE )
  {
    BITMAPINFOHEADER    stInfo;

    stInfo.cbFix = sizeof(BITMAPINFOHEADER);
    if ( GpiQueryBitmapParameters( pCC->hbmMem, &stInfo ) &&
         (stInfo.cx <= pRFBClient->width) &&
         (stInfo.cy <= pRFBClient->height) )
    {
      pRFBClient->width = stInfo.cx;
      pRFBClient->height = stInfo.cy;
      return TRUE;
    }
  }



  switch( pRFBClient->format.bitsPerPixel )
  {
    case 8:
    {
      /* libvncclient uses true-colors even for 8-bit bpp (indexed colors are
         not supported). There are no 8-bit true-color bitmap formats. So, we
         will use true-color bitmap, but with 16/32 bpp - what most closely
         matches the local desktop (to speed up). 8-bit bpp frame buffer data
         will be converted to 16/32 bpp bitmap at _cbRFBUpdate().  */
      LONG  alFormats[2];
      HPS   hpsDesktop = WinGetScreenPS( HWND_DESKTOP );

      ulBmpBPP = 16; // 16 bpp for bitmap by default.
      if ( hpsDesktop != NULLHANDLE )
      {
        if ( GpiQueryDeviceBitmapFormats( hpsDesktop, 2, (PLONG)&alFormats )
             && ( alFormats[1] == 32 ) )
          // Ok, local destop is 32 bpp, use it for bitmap too.
          ulBmpBPP = 32;
        WinReleasePS( hpsDesktop );
      }

      pRFBClient->format.redShift     = 0;
      pRFBClient->format.greenShift   = 3;
      pRFBClient->format.blueShift    = 6;
      pRFBClient->format.redMax       = 0x07;
      pRFBClient->format.greenMax     = 0x07;
      pRFBClient->format.blueMax      = 0x03;
      break;
    }

    case 16:
      ulBmpBPP = 16;
      pRFBClient->format.redShift     = 11;
      pRFBClient->format.greenShift   = 5;
      pRFBClient->format.blueShift    = 0;
      pRFBClient->format.redMax       = 0x1F;
      pRFBClient->format.greenMax     = 0x3F;
      pRFBClient->format.blueMax      = 0x1F;
      break;

    default: // 32
      ulBmpBPP = 32;
      pRFBClient->format.bitsPerPixel = 32;
      pRFBClient->format.redShift     = 16;
      pRFBClient->format.greenShift   = 8;
      pRFBClient->format.blueShift    = 0;
      pRFBClient->format.redMax       = 0xFF;
      pRFBClient->format.greenMax     = 0xFF;
      pRFBClient->format.blueMax      = 0xFF;
      break;
  }  // switch( pRFBClient->format.bitsPerPixel )
  
  if ( !SetFormatAndEncodings( pRFBClient ) )
  {
    debug( "SetFormatAndEncodings() failed" );
    return FALSE;
  }

  // Allocate frame buffer
  sizeAlloc = pRFBClient->width * pRFBClient->height * (ulBmpBPP>>3);
  if ( sizeAlloc >= SIZE_MAX )
  {
    rfbClientErr( "CRITICAL: cannot allocate frameBuffer, "
                  "requested size is too large\n" );
    return FALSE;
  }

  // Avoid access to bitmap: ccGetHPS() will locks caller.
  DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );

  if ( pRFBClient->frameBuffer != NULL )
    free( pRFBClient->frameBuffer );

  pRFBClient->frameBuffer = malloc( sizeAlloc );
  if ( pRFBClient->frameBuffer == NULL )
  {
    DosReleaseMutexSem( pCC->hmtxRFB );
    rfbClientErr( "CRITICAL: frameBuffer allocation failed, "
                  "requested size too large or not enough memory?\n" );
    return FALSE;
  }

  // Create memory bitmap.

  _ccDestoryBitmap( pCC );
  do
  {
    SIZEL    size;
    HDC      hdcMem = DevOpenDC( pCC->hab, OD_MEMORY, "*", 0, NULL, NULLHANDLE );

    // Create a new memory presentation space.
    size.cx = pRFBClient->width;
    size.cy = pRFBClient->height;
    pCC->hpsMem = GpiCreatePS( pCC->hab, hdcMem, &size,
                               PU_PELS | GPIF_DEFAULT | GPIT_MICRO | GPIA_ASSOC );
    if ( pCC->hpsMem == NULLHANDLE )
      debug( "GpiCreatePS() failed. Memory PS was not created." );
    else
    {
      PBITMAPINFOHEADER2 pbmih = alloca( /*ulBmpBPP == 8 // - always 16 or 32
                                      ? sizeof(BITMAPINFO2) + (sizeof(RGB2) << 8)
                                      :*/ sizeof(BITMAPINFO2) );
      PBITMAPINFO2       pbmi = NULL;

      if ( pbmih == NULL )
        debug( "Not enough stack size" );
      else
      {
        // Create a system bitmap object
        memset( pbmih, 0, sizeof(BITMAPINFOHEADER2) );
        pbmih->cbFix           = sizeof(BITMAPINFOHEADER2);
        pbmih->cx              = pRFBClient->width;
        pbmih->cy              = pRFBClient->height;
        pbmih->cPlanes         = 1;
        pbmih->cBitCount       = ulBmpBPP;

        pCC->hbmMem = GpiCreateBitmap( pCC->hpsMem, pbmih, 0, NULL, pbmi );
        if ( ( pCC->hbmMem == GPI_ERROR ) || ( pCC->hbmMem == NULLHANDLE ) )
        {
          debug( "GpiCreateBitmap() failed: %u x %u, bpp: %u",
                 pbmih->cx, pbmih->cy, ulBmpBPP );
          pCC->hbmMem = NULLHANDLE;
        }
        // Set bitmap object for the memory presentation space.
        else if ( GpiSetBitmap( pCC->hpsMem, pCC->hbmMem ) == HBM_ERROR )
          debug( "GpiSetBitmap() failed" );
        else
          break; // Success.
      }
    }  // if ( pCC->hpsMem == NULLHANDLE ) else

    // Error.
    _ccDestoryBitmap( pCC );
    DosReleaseMutexSem( pCC->hmtxRFB );
    rfbClientErr( "CRITICAL: frame bitmap creation failed.\n" );
    return FALSE;
  }
  while( FALSE );

  // Set window size (the window may not be set at this time).

  if ( pCC->hwnd == NULLHANDLE )
    debug( "Window is not yet set" );
  else
    WinPostMsg( pCC->hwnd, WM_VNC_SETCLIENTSIZE,
                MPFROM2SHORT( pCC->pRFBClient->width,
                              pCC->pRFBClient->height ), 0 );

  DosReleaseMutexSem( pCC->hmtxRFB );

  return TRUE;
}

static void _cbRFBUpdate(rfbClient* pRFBClient, int x, int y, int w, int h)
{
  PCLNTCONN        pCC = (PCLNTCONN)rfbClientGetClientData( pRFBClient, NULL );
  struct {
    BITMAPINFOHEADER2  bmp2;
    RGB2               argb2Color[0x100];
  }                bm;
  ULONG            ulIdx;
  ULONG            cbLine = pRFBClient->width *
                            ( pRFBClient->format.bitsPerPixel >> 3 );
  PBYTE            pbData;
  LONG             lRC;
  HPS              hpsMem = ccGetHPS( pCC );
  LONG             lYTop;

  // Copy chaged lines from the frame buffer to the bitmap.

  memset( &bm, 0, sizeof(bm) );
  bm.bmp2.cbFix = sizeof(BITMAPINFOHEADER2);//16;
  if ( !GpiQueryBitmapInfoHeader( pCC->hbmMem, &bm.bmp2 ) )
  {
    debug( "GpiQueryBitmapInfoHeader() failed" );
    return;
  }

  pbData = (PBYTE)&pRFBClient->frameBuffer[y * cbLine];
  lYTop = bm.bmp2.cy - y - 1; // Y - top line relative bottom of bitmap.

  if ( pRFBClient->format.bitsPerPixel == 8 )
  {
    PBYTE    pbLine = malloc( ((bm.bmp2.cBitCount * bm.bmp2.cx) + 1) & ~0x01 );

    for( ulIdx = 0; ulIdx < h; ulIdx++, lYTop-- )
    {
      if ( bm.bmp2.cBitCount == 16 )
        _conv8to16( pbLine, pbData, bm.bmp2.cx );
      else // 32
        _conv8to32( pbLine, pbData, bm.bmp2.cx );

      lRC = GpiSetBitmapBits( pCC->hpsMem, lYTop, 1, pbLine,
                              (PBITMAPINFO2)&bm );
      if ( lRC == 0 || lRC == GPI_ALTERROR )
        debug( "GpiSetBitmapBits(,%u,,,) failed", lYTop );
      pbData += cbLine;
    }

    free( pbLine );
  }
  else
    for( ulIdx = 0; ulIdx < h; ulIdx++, lYTop-- )
    {
      lRC = GpiSetBitmapBits( pCC->hpsMem, lYTop, 1, pbData,
                              (PBITMAPINFO2)&bm );
      if ( lRC == 0 || lRC == GPI_ALTERROR )
        debug( "GpiSetBitmapBits(,%u,,,) failed", lYTop );
      pbData += cbLine;
    }

  ccReleaseHPS( pCC, hpsMem );

  if ( pCC->hwnd != NULLHANDLE )
  {
#if 0
    RECTL    rect;

    rect.xLeft = x;
    rect.yTop = bm.bmp2.cy - y;
    rect.xRight = rect.xLeft + w;
    rect.yBottom = rect.yTop - h;

    if ( !WinInvalidateRect( pCC->hwnd, &rect, FALSE ) )
      debug( "WinInvalidateRect() failed" );
#else
    SHORT sXLeft = x;
    SHORT sYTop = bm.bmp2.cy - y;
    SHORT sXRight = sXLeft + w;
    SHORT sYBottom = sYTop - h;

    WinPostMsg( pCC->hwnd, WM_VNC_UPDATE,
                MPFROM2SHORT(sXLeft, sYBottom), MPFROM2SHORT(sXRight, sYTop) );
#endif
  }
  else
    debugPCP( "No window to send WM_VNC_UPDATE." );
}

static void _cbRFBKeyboardLedState(rfbClient* cl, int value, int pad)
{
  /* note: pad is for future expansion 0=unused */
  debugPCP();
}

static void _cbRFBTextChat(rfbClient* pRFBClient, int value, char *text)
{
  PCLNTCONN        pCC = (PCLNTCONN)rfbClientGetClientData( pRFBClient, NULL );

  if ( pCC->hwnd == NULLHANDLE )
  {
    debug( "Window is not yet set" );
    return;
  }

  switch( value )
  {
    case rfbTextChatOpen:
      debug( "TextChat: We should open a textchat window" );
      WinSendMsg( pCC->hwnd, WM_VNC_CHAT, MPFROMLONG(CCCHAT_OPEN), 0 );
      TextChatOpen( pRFBClient );
      break;

    case rfbTextChatClose:
    case rfbTextChatFinished:
      debug( "TextChat: We should close our window" );
      WinSendMsg( pCC->hwnd, WM_VNC_CHAT, MPFROMLONG(CCCHAT_CLOSE), 0 );
      break;

    default:
      {
        PSZ      pszEncText = ccLRConvNew( pCC->pRFBClient, CNV_REMOTETOLOCAL,
                                           -1, text, NULL );

        debug( "TextChat: Received \"%s\" (\"%s\")", text, pszEncText );

        WinSendMsg( pCC->hwnd, WM_VNC_CHAT, MPFROMLONG(CCCHAT_MESSAGE),
                    MPFROMP(pszEncText == NULL ? text : pszEncText) );

        if ( pszEncText != NULL )
          free( pszEncText );
      }
      break;
  }
}

static char* _cbRFBGetPassword(rfbClient *pRFBClient)
{
  PCLNTCONN        pCC = (PCLNTCONN)rfbClientGetClientData( pRFBClient, NULL );
  CCLOGONINFO      stLogonInfo;
  PSZ              pszPassword;

  debugCP();
  stLogonInfo.fCredential = FALSE;
  if ( !_logonInfoToThread( pCC, &stLogonInfo ) )
  {
    debugCP( "_logonInfoToThread(), rc = FALSE" );
    return NULL;
  }

#ifdef DEBUG_CODE
// We should use "original" strdup() and malloc() functions. Memory allocated
// here for the password will be freed in libvncclient library (it's so ugly!).
// DEBUG_CODE was defined in debug.h.
#undef strdup
#undef malloc
#endif
  if ( stLogonInfo.acPassword[0] == '\0' )
  {
    // We can get a blank password. Function HandleVncAuth() in
    // libvncclient\rfbproto.c has appropriate changes.
    pszPassword = malloc( 1 );
    pszPassword[0] = '\0';
  }
  else
    pszPassword = strdup( stLogonInfo.acPassword );
#ifdef DEBUG_CODE
// Continue to use our debug-related routines...
#define strdup(src) debug_strdup(src, __FILE__, __LINE__)
#define malloc(size) debug_malloc(size, __FILE__, __LINE__)
#endif
  memset( &stLogonInfo, 0, sizeof(CCLOGONINFO) );

  return pszPassword;
}

static rfbCredential* _cbRFBGetCredential(rfbClient *pRFBClient,
                                          int credentialType)
{
  PCLNTCONN        pCC = (PCLNTCONN)rfbClientGetClientData( pRFBClient, NULL );
  CCLOGONINFO      stLogonInfo;
  rfbCredential    *pRFBCred;

  debugCP();

  /* credentialType = rfbCredentialTypeX509
     rfbCredential->x509Credential
       x509CACertFile
       x509CACrlFile
       x509ClientCertFile
       x509ClientKeyFile
       x509CrlVerifyMode
  */
  if ( credentialType == rfbCredentialTypeX509 )
  {
    debugCP( "credentialType == rfbCredentialTypeX509" );
#ifdef DEBUG_CODE
#undef calloc
#undef malloc
#undef strdup
#endif
    pRFBCred = calloc( 1,sizeof(rfbCredential) );
    pRFBCred->x509Credential.x509CrlVerifyMode = rfbX509CrlVerifyNone;
#ifdef DEBUG_CODE
#define calloc(n,size) debug_calloc(n, size, __FILE__, __LINE__)
#define malloc(size) debug_malloc(size, __FILE__, __LINE__)
#define strdup(src) debug_strdup(src, __FILE__, __LINE__)
#endif
    return pRFBCred;
  }

  if ( credentialType != rfbCredentialTypeUser ) // rfbCredentialTypeX509
  {
    debug( "Unsupported logon information type requested" );
    return NULL;
  }

  stLogonInfo.fCredential = TRUE;
  if ( !_logonInfoToThread( pCC, &stLogonInfo ) )
    return NULL;

#ifdef DEBUG_CODE
#undef malloc
#undef strdup
#endif
  pRFBCred = malloc( sizeof(rfbCredential) );
  pRFBCred->userCredential.username = strdup( stLogonInfo.acUserName );
  pRFBCred->userCredential.password = strdup( stLogonInfo.acPassword );
#ifdef DEBUG_CODE
#define malloc(size) debug_malloc(size, __FILE__, __LINE__)
#define strdup(src) debug_strdup(src, __FILE__, __LINE__)
#endif
  memset( &stLogonInfo, 0, sizeof(CCLOGONINFO) );

  return pRFBCred;
}

// We support cursors for 16 and 32 BPP. It seems, libvncclient don't give me
// color cursor image for 8 BPP.
static void cbGotCursorShape(rfbClient *pRFBClient, int xhot, int yhot,
                             int width, int height, int bytesPerPixel)
{
  PCLNTCONN            pCC = (PCLNTCONN)rfbClientGetClientData( pRFBClient, NULL );
  ULONG                ulBPP, ulRC;
  PBITMAPINFOHEADER2   pbmih;
  PBITMAPINFO2         pbmi = NULL;
  PBYTE                pbImage;
  PBYTE                pbLine;
  ULONG                cbSrcLine, cbDstLine;
  PBYTE                pbInMask = (PBYTE)&pRFBClient->rcMask[ (width - 1) *
                                                              height ];
                                  // Last line in source mask.
  PBYTE                pbInImage, pbInMaskLine;
  HPS                  hps;
  register ULONG       ulPos;
  ULONG                ulLine;
  POINTERINFO          ptri = { 0 };

  if ( pCC->pRFBClient->appData.viewOnly )
    // Do not set cursor in view-only mode.
    return;

  pbmih = alloca( sizeof(BITMAPINFO2) + (2*sizeof(RGB2)) );
  if ( pbmih == NULL )
  {
    debug( "Not enough stack size" );
    return;
  }

  memset( pbmih, 0, sizeof(BITMAPINFOHEADER2) );
  pbmi           = (PBITMAPINFO2)pbmih;

  // Create cursor image bitmap.
  // -------------------------

/*  if ( bytesPerPixel == 1 )
    ulBPP = 16;
  else*/
    ulBPP = bytesPerPixel << 3;

  cbSrcLine = bytesPerPixel * width;
  cbDstLine = ( (ulBPP * width + 31) / 32 ) * 4;

  pbmih->cbImage = cbDstLine * height;

  // Buffer to store converted data for the system bitmap.
  pbImage = alloca( pbmih->cbImage );
  if ( pbImage == NULL )
  {
    debug( "Not enough stack size" );
    return;
  }

  pbInImage = (PBYTE)&pRFBClient->rcSource[pbmih->cbImage - cbSrcLine]; // Last line in source.
  pbInMaskLine = pbInMask; // Last line in source mask.

  for( ulLine = 0, pbLine = pbImage; ulLine < height;
       ulLine++, pbLine += cbDstLine,
       pbInImage -= cbSrcLine, pbInMaskLine -= width )
  {
/*    if ( bytesPerPixel == 1 )
      _conv8to16( pbLine, pbInMaskLine, width );
    else*/

    if ( bytesPerPixel == 2 )
    {
      for( ulPos = 0; ulPos < width; ulPos++ )
        ((PUSHORT)pbLine)[ulPos] = pbInMaskLine[ulPos] == 0
                                          ? 0 : ((PUSHORT)pbInImage)[ulPos];
    }
    else
    {
      for( ulPos = 0; ulPos < width; ulPos++ )
        ((PULONG)pbLine)[ulPos] = pbInMaskLine[ulPos] == 0
                                          ? 0 : ((PULONG)pbInImage)[ulPos];
    }

    // Padding.
    memset( &pbLine[cbSrcLine], 0, cbDstLine - cbSrcLine );
  }

  // Create a system bitmap object
  pbmih->cbFix     = sizeof(BITMAPINFOHEADER2);
  pbmih->cx        = width;
  pbmih->cy        = height;
  pbmih->cPlanes   = 1;
  pbmih->cBitCount = ulBPP;

  hps = WinGetPS( pCC->hwnd );
  ptri.hbmColor = GpiCreateBitmap( hps, pbmih, CBM_INIT, pbImage, pbmi );

  if ( ( ptri.hbmColor == GPI_ERROR ) || ( ptri.hbmColor == 0 ) )
  {
    debug( "GpiCreateBitmap() failed" );
    WinReleasePS( hps );
    return;
  }

  // Create cursor mask bitmap
  // -------------------------

  // AND mask.
  cbDstLine = ( (width + 31) / 32 ) * 4;
  memset( pbImage, 0x00, height * cbDstLine  );

  // XOR mask, convert from input.
  pbLine = &pbImage[height * cbDstLine];
  for( ulLine = 0, pbInMaskLine = pbInMask; ulLine < height;
       ulLine++, pbLine += cbDstLine, pbInMaskLine -= width )
  {
    memset( pbLine, 0, cbDstLine );
    for( ulPos = 0; ulPos < width; ulPos++ )
    {
      if ( pbInMaskLine[ulPos] == 0 )
        pbLine[ulPos >> 3] |= 0x80 >> (ulPos & 7);
    }
  }

  *((PULONG)&pbmi->argbColor[0]) = 0x00000000;
  *((PULONG)&pbmi->argbColor[1]) = 0x00FFFFFF;

  // Create a system bitmap object
  pbmih->cy        = 2 * height; // Doubhe height: invert mask and source mask.
  pbmih->cBitCount = 1;
  pbmih->cbImage   = cbDstLine * height;

  ptri.hbmPointer = GpiCreateBitmap( hps, pbmih, CBM_INIT, pbImage, pbmi );
  WinReleasePS( hps );

  if ( ( ptri.hbmPointer == GPI_ERROR ) || ( ptri.hbmPointer == 0 ) )
  {
    debug( "GpiCreateBitmap() failed" );
    return;
  }

  // Create system pointer
  // ---------------------

  ptri.fPointer = TRUE;
  ptri.xHotspot = xhot;
  ptri.yHotspot = height - yhot - 1;

  ulRC = DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );
  if ( ulRC != NO_ERROR )
    debug( "DosRequestMutexSem(), rc = ", ulRC );

  if ( pCC->hptrPointer != NULLHANDLE )
    WinDestroyPointer( pCC->hptrPointer );

  pCC->hptrPointer = WinCreatePointerIndirect( HWND_DESKTOP, &ptri );
  if ( pCC->hptrPointer == NULLHANDLE )
    debug( "WinCreatePointerIndirect() failed" );

  DosReleaseMutexSem( pCC->hmtxRFB );

  // Destroy bitmaps.
  GpiDeleteBitmap( ptri.hbmPointer );
  GpiDeleteBitmap( ptri.hbmColor );
}

static void cbGotXCutText(rfbClient *pRFBClient, const char *pcText, int cbText)
{
  PCLNTCONN  pCC = (PCLNTCONN)rfbClientGetClientData( pRFBClient, NULL );
  CWCBDATA   stData;

  debug( "Text: \"%s\"", pcText );
  if ( ( pcText == NULL ) || ( pCC->hwnd == NULLHANDLE ) )
    return;

  stData.cbTextUnicode = 0;
  stData.pcTextUnicode = NULL;
  stData.cbText = cbText;
  stData.pcText = (PCHAR)pcText;
  WinSendMsg( pCC->hwnd, WM_VNC_CLIPBOARD, MPFROMP(&stData), 0 );
}

static void cbGotXCutTextUTF8(rfbClient *pRFBClient, const char *pcText,
                              int cbText)
{
  PCLNTCONN  pCC = (PCLNTCONN)rfbClientGetClientData( pRFBClient, NULL );
  CWCBDATA   stData = { 0 };

  if ( ( pcText == NULL ) || ( pCC->hwnd == NULLHANDLE ) )
    return;

  stData.pcTextUnicode = cnvConvNew( "UTF-16LE", "UTF-8",
                                     cbText, (PCHAR)pcText,
                                     (size_t *)&stData.cbTextUnicode );
  stData.pcText = cnvConvNew( "", "UTF-8", cbText, (PCHAR)pcText,
                              (size_t *)&stData.cbText );

  WinSendMsg( pCC->hwnd, WM_VNC_CLIPBOARD, MPFROMP(&stData), 0 );

  if ( stData.pcTextUnicode != NULL )
    free( stData.pcTextUnicode );
  if ( stData.pcText != NULL )
    free( stData.pcText );
}

static void cbBell(rfbClient *pRFBClient)
{
  WinAlarm( HWND_DESKTOP, WA_NOTE );
}

// Thread.

static BOOL _initConnection(rfbClient *pRFBClient)
{
  PLRCONV  pLRConv = (PLRCONV)rfbClientGetClientData( pRFBClient, _TAG_LRCONV );
  char     *listenAddress = pRFBClient->listenAddress;
  uint8_t  *frameBuffer = pRFBClient->frameBuffer;
  char     *encodingsString = (char *)pRFBClient->appData.encodingsString;

  debugCP( "rfbInitClient()..." );
  if ( rfbInitClient( pRFBClient, NULL, NULL ) )
  {
    debugCP( "Success" );
    return TRUE;
  }
  debugCP( "Fail" );

  /* If it fails, the rfbInitClient() function destroys the rfbClient object,
     but we still need to destroy our stuff.  */

  cnvLRDestroy( pLRConv );

  if ( listenAddress != NULL )
    free( listenAddress );

  if ( frameBuffer != NULL )
    free( frameBuffer );

  if ( encodingsString != NULL )
    free( encodingsString );

  return FALSE;
}

static void threadClient(void *pArg)
{
  PCLNTCONN  pCC = (PCLNTCONN)pArg;
  int        iRC;
  HMQ        hmq;

  _ccSetThreadState( pCC, RFBSTATE_INITPROGRESS );

  // PM stuff initialization.
  pCC->hab = WinInitialize( 0 );
  hmq = WinCreateMsgQueue( pCC->hab, 0 );
  if ( hmq == NULLHANDLE )
  {
    debug( "WinCreateMsgQueue() failed" );
    _ccSetThreadState( pCC, RFBSTATE_FINISH );
    _endthread();
  }

  if ( pCC->pRFBClient->listenPort >= 0 )
    listenForIncomingConnectionsNoFork( pCC->pRFBClient, -1 );

  if ( !_initConnection( pCC->pRFBClient ) )
  {
    debug( "rfbInitConnection() failed" );
    pCC->pRFBClient = NULL;
  }
  else
  {
    ULONG  ulRC = DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );

    if ( ulRC != NO_ERROR )
      debug( "#1 DosRequestMutexSem(), rc = %lu", ulRC );

    if ( ( ulRC == NO_ERROR ) && ( pCC->ulStopSigTime == 0 ) )
    {
      rfbClientLog( "VNC client format: %u bits per pixel.\n",
                    pCC->pRFBClient->format.bitsPerPixel );

      _ccSetThreadState( pCC, RFBSTATE_READY );

      while( pCC->ulStopSigTime == 0 )
      {
        DosReleaseMutexSem( pCC->hmtxRFB );

        iRC = fxWaitForMessage( pCC->pRFBClient, 50000 );
        if ( iRC < 0 )
        {
          debug( "fxWaitForMessage(), rc = %d", iRC );
          ulRC = ~0;  // mutex released
          break;
        }

        if ( iRC != 0 )
        {
          while( HandleRFBServerMessage( pCC->pRFBClient ) &&
                 pCC->pRFBClient->buffered && ( pCC->ulStopSigTime == 0 ) )
            { }

          if ( recv( pCC->pRFBClient->sock, &iRC, 1, MSG_PEEK ) == 0 )
          {
            debugPCP( "Connection closed" );
            ulRC = ~0;  // mutex released
            break;
          }
        }

        ulRC = DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );
        if ( ulRC != NO_ERROR )
        {
          debug( "#2 DosRequestMutexSem(), rc = %lu", ulRC );
          break;
        }

        if ( pCC->pRFBKeySeq != NULL )
        {
          ULONG  ulIndex = pCC->pRFBKeySeq->ulIndex;

          if ( ulIndex == pCC->pRFBKeySeq->ulCount )
          {
            free( pCC->pRFBKeySeq );
            pCC->pRFBKeySeq = NULL;
          }
          else
          {
            ULONG  ulKey = pCC->pRFBKeySeq->aulKeys[ulIndex];

            ccSendKeyEvent( pCC, ulKey & ~CCKEY_PRESS, (ulKey & CCKEY_PRESS)!=0 );
            pCC->pRFBKeySeq->ulIndex++;
          }
        }
      }  // while( pCC->ulStopSigTime == 0 )

      debug( "End thread cycle..." );
    }  // if ( ( ulRC == NO_ERROR ) && ( pCC->ulStopSigTime == 0 ) )

    if ( ulRC == NO_ERROR )
      DosReleaseMutexSem( pCC->hmtxRFB );

  }  // if ( !_initConnection( pCC->pRFBClient ) ) else

  _ccSetThreadState( pCC, RFBSTATE_FINISH );
  WinDestroyMsgQueue( hmq );
  debugCP( "End thread" );

  _endthread();
}

// Global log handle.

static void _cbLibLog(const char *format, ...)
{
  PTIB       tib;
  PPIB       pib;
  PCLNTCONN  pCC;
  va_list    args;
  int        cbBuf;
  CHAR       acBuf[512];

  // Ignore chat messages.
  if ( strncmp( format, "Received TextChat ",
                strlen( "Received TextChat " ) ) == 0 )
    return;

  va_start(args, format);
  cbBuf = _vsnprintf( acBuf, sizeof(acBuf), format, args );
  va_end(args);

  if ( cbBuf == -1 )
  {
    debugCP( "WTF?!" );
    return;
  }

  if ( cbBuf > sizeof(acBuf) )
    cbBuf = sizeof(acBuf);
  printf( "%.*s", cbBuf, acBuf );

  DosGetInfoBlocks( &tib, &pib );
  DosRequestMutexSem( hmtxCCList, SEM_INDEFINITE_WAIT );

  // Search CLNTCONN.

  for( pCC = (PCLNTCONN)lnkseqGetFirst( &lsCCList ); pCC != NULL;
       pCC = (PCLNTCONN)lnkseqGetNext( pCC ) )
  {
    if ( pCC->iRFBTid == tib->tib_ptib2->tib2_ultid )
      break;
  }

  if ( pCC == NULL )
    debug( "Thread for the new log record is not found. Text: %s", acBuf );
  else
  {
    // Create a log record.

#ifdef _LASTLOGENTRYONLY
    while( ( cbBuf != 0 ) && isspace( acBuf[cbBuf-1] ) )
      cbBuf--;

    DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );

    if ( cbBuf >= pCC->cbLastLogEntMax )
    {
      if ( pCC->pcLastLogEnt != NULL )
        free( pCC->pcLastLogEnt );

      pCC->pcLastLogEnt = malloc( cbBuf + 1 /* +1 - trailing ZERO */ );
      pCC->cbLastLogEntMax = pCC->pcLastLogEnt == NULL ? 0 : (cbBuf + 1);
    }

    if ( pCC->pcLastLogEnt != NULL )
    {
      memcpy( pCC->pcLastLogEnt, acBuf, cbBuf );
      pCC->pcLastLogEnt[cbBuf] = '\0';
      debug( "Log record stored: %.*s", cbBuf, acBuf );
    }

    DosReleaseMutexSem( pCC->hmtxRFB );
#else
    PLOGREC  pLogRec = malloc( sizeof(LOGREC) + cbBuf );

    if ( pLogRec != NULL )
    {
      time( &pLogRec->timeRec );
      memcpy( pLogRec->acText, acBuf, cbBuf );
      pLogRec->acText[cbBuf] = '\0';

      DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );
      lnkseqAddFirst( &pCC->lsLog, pLogRec );
      DosReleaseMutexSem( pCC->hmtxRFB );
      debug( "Log record stored: %s", acBuf );
    }
#endif
  }

  DosReleaseMutexSem( hmtxCCList );
}

static VOID _cbKbdMapLoadErr(PSZ pszFile, ULONG ulLine, ULONG ulCode)
{
  PSZ        pszMsg;

  switch( ulCode )
  {
    case XKEYERR_KEYSYM:
      pszMsg = "Invalid keysym code";
      break;

    case XKEYERR_CHAR:
      pszMsg = "Invalid characted code";
      break;

    case XKEYERR_FLAG:
      pszMsg = "Unknown flag(s)";
      break;

    case XKEYERR_SCAN:
      pszMsg = "Invalid scan code";
      break;

    case XKEYERR_VK:
      pszMsg = "Unknown virtual key name";
      break;

    default:
      printf( "Unknown error code 0x%X\n", ulCode );
      return;
  }

  printf( "%s at line %u in file %s\n", pszMsg, ulLine, pszFile );
}


/*
 *  Need the source code of UltraVNC SecureVNC Plugin...
 *

#define rfbUltraVNC_SecureVNCPluginAuth_new 0x73

rfbBool HandleAnonTLSAuth(rfbClient* client);

static uint32_t aui32UltraVNCSec[1] = { rfbUltraVNC_SecureVNCPluginAuth_new };

static rfbBool _UltraVNC_SecureVNCPluginAuth_new_Auth(rfbClient* client, uint32_t authScheme)
{
  uint16_t  cbChallenge;
  uint8_t   *pChallenge;
  uint8_t   ui8Passphraseused;

//  HandleAnonTLSAuth( client );

  if ( !ReadFromRFBServer( client, (char *)&cbChallenge, 2 ) )
    return FALSE;

//  cbChallenge = rfbClientSwap16IfLE( cbChallenge );
  printf( "Challenge length: %u\n", cbChallenge );

  pChallenge = malloc( cbChallenge );
  if ( !ReadFromRFBServer( client, (char *)pChallenge, cbChallenge ) )
    return FALSE;

  puts( "Challenge readed" );

  // UltraVNC vncviewer\ClientConnection.cpp  Line 3159
  if ( !ReadFromRFBServer( client, (char *)&ui8Passphraseused, 1 ) )
    return FALSE;

  printf( "ui8Passphraseused: %u\n", ui8Passphraseused );

  . . .

  return FALSE;
}

rfbClientProtocolExtension stUltraVNCProtocolExtension = {
  encodings: NULL,
  handleEncoding: NULL,
  handleMessage: NULL,
  next: NULL,
  securityTypes: aui32UltraVNCSec,
  handleAuthentication: _UltraVNC_SecureVNCPluginAuth_new_Auth
};
*/

// Public routines.

PSZ ccLRConvNew(rfbClient *pRFBClient, ULONG ulDirection /* CNV_xxxxx */,
                LONG cbStr, PCHAR pcStr, PULONG pcbLength)
{
  PLRCONV  pLRConv = (PLRCONV)rfbClientGetClientData( pRFBClient, _TAG_LRCONV );

  return (PSZ)cnvLRConvNew( pLRConv, ulDirection, cbStr, pcStr,
                            (size_t *)pcbLength );
}

ULONG ccInit()
{
  ULONG      ulRC;

  if ( hmtxCCList != NULLHANDLE )
  {
    debug( "Already initialized" );
    return CC_INITIALIZED;
  }

  ulRC = DosCreateMutexSem( NULL, &hmtxCCList, 0, FALSE );
  if ( ulRC != NO_ERROR )
  {
    debug( "DosCreateMutexSem(), rc = %u", ulRC );
    return CC_SEMERROR;
  }

  pKbdMap = xkMapLoad( "keysym\\user.xk & keysym\\%L.xk & keysym\\general.xk",
                       _cbKbdMapLoadErr );
  if ( pKbdMap == NULL )
  {
    DosCloseMutexSem( hmtxCCList );
    hmtxCCList = NULLHANDLE;
    return CC_NOKBDMAP;
  }

  lnkseqInit( &lsCCList );

  rfbClientLog = _cbLibLog;
  rfbClientErr = _cbLibLog;

  fxRegister();
//  rfbClientRegisterExtension( &stUltraVNCProtocolExtension );

  return CC_OK;
}

// Destroy all connection objects and windows.
VOID ccDone()
{
  HWND       hwnd;

  debugCP();

  if ( hmtxCCList == NULLHANDLE )
  {
    debug( "Not initialized" );
    return;
  }

  if ( lnkseqGetCount( &lsCCList ) != 0 )
  {
    debug( "%u objects was not destroyed, do it now...",
           lnkseqGetCount( &lsCCList ) );
  }

  while( lnkseqGetFirst( &lsCCList ) != NULL )
  {
    hwnd = ((PCLNTCONN)lnkseqGetFirst( &lsCCList ))->hwnd;

    ccDestroy( (PCLNTCONN)lnkseqGetFirst( &lsCCList ) );
    if ( hwnd != NULLHANDLE )
      WinDestroyWindow( hwnd );
  }

  DosCloseMutexSem( hmtxCCList );
  hmtxCCList = NULLHANDLE;

  xkMapFree( pKbdMap );
}

PCLNTCONN ccCreate(PCCPROPERTIES pProperties, HWND hwnd)
{
  ULONG       ulRC;
  PCLNTCONN   pCC = calloc( 1, sizeof(CLNTCONN) );
  PLRCONV     pLRConv = NULL;
  rfbClient*  pRFBClient;

  if ( pCC == NULL )
    return NULL;

  if ( !pProperties->fViewOnly )
  {
    pCC->hptrPointer = WinLoadPointer( HWND_DESKTOP, NULLHANDLE, IDPTR_DEFPTR );
    if ( pCC->hptrPointer == NULLHANDLE )
      debug( "WinLoadPointer() failed" );
  }

  pCC->hwnd = hwnd;
#ifndef _LASTLOGENTRYONLY
  lnkseqInit( &pCC->lsLog );
#endif
  xkKeysymFromMPStart( &pCC->stXKFromMP );
  pCC->iRFBTid = -1;

  do
  {
    ulRC = DosCreateMutexSem( NULL, &pCC->hmtxRFB, 0, FALSE );
    if ( ulRC != NO_ERROR )
    {
      debug( "DosCreateMutexSem(), rc = %u", ulRC );
      break;
    }

    switch( pProperties->ulBPP )
    {
      case 8:
        pRFBClient = rfbGetClient( 8, 1, 1 ); // 8 BPP
        break;

      case 16:
        pRFBClient = rfbGetClient( 8, 2, 2 ); // 16 BPP
        break;

      default:
        pRFBClient = rfbGetClient( 8, 3, 4 ); // 32 BPP
    }

    if ( pRFBClient == NULL )
    {
      debug( "rfbGetClient() - result is NULL" );
      break;
    }
    pCC->pRFBClient = pRFBClient;

    pRFBClient->MallocFrameBuffer       = _cbRFBResize;
    pRFBClient->canHandleNewFBSize      = TRUE;
    pRFBClient->GotFrameBufferUpdate    = _cbRFBUpdate;
    pRFBClient->HandleKeyboardLedState  = _cbRFBKeyboardLedState;
    pRFBClient->HandleTextChat          = _cbRFBTextChat;
    pRFBClient->listenPort              = pProperties->lListenPort;
    pRFBClient->listen6Port             = pProperties->lListenPort;
    pRFBClient->GetPassword             = _cbRFBGetPassword;
    pRFBClient->GetCredential           = _cbRFBGetCredential;
    pRFBClient->GotCursorShape          = cbGotCursorShape;
    pRFBClient->GotXCutText             = cbGotXCutText;
    pRFBClient->GotXCutTextUTF8         = cbGotXCutTextUTF8;
    pRFBClient->Bell                    = cbBell;
    pRFBClient->appData.useBGR233       = TRUE;
    // We don't support cursor for 8 BPP.
    pRFBClient->appData.useRemoteCursor = pProperties->ulBPP != 8;
    pRFBClient->appData.viewOnly        = pProperties->fViewOnly;
    pRFBClient->appData.shareDesktop    = pProperties->fShareDesktop;

    pRFBClient->programName             = APP_NAME;
    pRFBClient->appData.encodingsString =
                                pProperties->acEncodings[0] == '\0'
                                  ? NULL : strdup( pProperties->acEncodings );
    if ( pProperties->ulCompressLevel <= 9 )
      pRFBClient->appData.compressLevel = pProperties->ulCompressLevel;
    if ( pProperties->ulQualityLevel <= 9 )
      pRFBClient->appData.qualityLevel = pProperties->ulQualityLevel;
    if ( pProperties->ulQoS_DSCP != 0 )
      pRFBClient->QoS_DSCP = (int)pProperties->ulQoS_DSCP;

    if ( pProperties->acDestHost[0] != '\0' )
    {
      PCHAR  pcColon = strchr( pProperties->acDestHost, ':' );

      pRFBClient->destHost = strdup( pProperties->acDestHost );
      if ( ( pRFBClient->destHost != NULL ) && ( pcColon != NULL ) )
      {
        pRFBClient->destHost[(int)(pcColon - pProperties->acDestHost)] = '\0';
        pRFBClient->destPort = atoi( &pcColon[1] );
      }
    }

    if ( pProperties->lListenPort < 0 )
    {
      PCHAR  pcColon = strrchr( pProperties->acHost, ':' );

#ifdef DEBUG_CODE
/* We should use "original" strdup() and free() functions. Memory allocated
   here will be freed in libvncclient library DEBUG_CODE was defined in
   debug.h.  */
#undef strdup
#undef free
#endif
      if ( pRFBClient->serverHost != NULL )
        free( pRFBClient->serverHost );
      pRFBClient->serverHost = strdup( pProperties->acHost );
#ifdef DEBUG_CODE
// Continue to use our debug-related routines...
#define strdup(src) debug_strdup(src, __FILE__, __LINE__)
#define free(ptr) debug_free(ptr, __FILE__, __LINE__)
#endif
      if ( ( pRFBClient->serverHost != NULL ) && ( pcColon != NULL ) )
      {
        pRFBClient->serverHost[pcColon - pProperties->acHost] = '\0';
        pRFBClient->serverPort = atoi( &pcColon[1] );
      }

      // Accept values 0..9 as "display", otherwise - port number.
      if ( ( pRFBClient->serverPort >= 0 ) && ( pRFBClient->serverPort <= 9 ) )
        pRFBClient->serverPort += 5900;
    }  // if ( pProperties->lListenPort < 0 )

    if ( pProperties->lListenPort >= 0 )
    {
      // "Listen mode", set local interface.
      pCC->pRFBClient->listenAddress = pProperties->acHost[0] == '\0' ?
                                         NULL : strdup( pProperties->acHost );
    }

    if ( pProperties->fNoTLS )
    {
      // Allow all security authentication schemes except those supporting TLS.
      const uint32_t  aSchemes[] = { rfbNoAuth, rfbVncAuth, rfbTight,
                                     rfbUltra, /*rfbVeNCrypt,*/ rfbSASL, rfbARD,
                                     rfbUltraMSLogonII/*, rfbTLS*/ };
      SetClientAuthSchemes( pCC->pRFBClient, aSchemes, ARRAYSIZE(aSchemes) );
    }
    rfbClientSetClientData( pCC->pRFBClient, NULL, pCC );

    pCC->pRFBClient->appData.scaleSetting = ( pProperties->ulScale == 0 ) ||
                                            ( pProperties->ulScale > 10 )
                                              ? 1 : pProperties->ulScale;

    if ( pProperties->acCharset[0] != '\0' )
    {
      pLRConv = cnvLRCreate( pProperties->acCharset );
      if ( pLRConv == NULL )
        rfbClientLog( "Failed to open character encoding: %s\n",
                      pProperties->acCharset );
    }
    rfbClientSetClientData( pCC->pRFBClient, _TAG_LRCONV, pLRConv );

    // Run thread.
    pCC->ulRFBState = RFBSTATE_NOTINIT;
    pCC->iRFBTid = _beginthread( threadClient, NULL, 65535, pCC );
    if ( pCC->iRFBTid == -1 )
    {
      debug( "_beginthread() failed" );
      break;
    }

    // Store pointer to the new object at the list of objests.
    debugCP( "Store pointer to the new object at the list of objests" );
    ulRC = DosRequestMutexSem( hmtxCCList, SEM_INDEFINITE_WAIT );
    if ( ulRC != NO_ERROR )
      debug( "DosRequestMutexSem(), rc = %u", ulRC );
    lnkseqAdd( &lsCCList, pCC );
    debug( "Client connection objects: %u", lnkseqGetCount( &lsCCList ) );
    DosReleaseMutexSem( hmtxCCList );

    // Success.
    return pCC;
  }
  while( FALSE );

  // Client connection object creation failed,

  if ( pCC != NULL )
  {
    if ( pCC->hmtxRFB != NULLHANDLE )
      DosCloseMutexSem( pCC->hmtxRFB );

    _RFBClientCleanUp( pCC->pRFBClient );
    free( pCC );
  }

  return NULL;
}

VOID ccDestroy(PCLNTCONN pCC)
{
  ULONG      ulRC;
  int        iRFBTid;

  debugCP( "Enter" );
  if ( pCC == NULL )
  {
    debugCP( "Argument is NULL. Return." );
    return;
  }

  // Remove pointer to the object from the list of objests.
  ulRC = DosRequestMutexSem( hmtxCCList, SEM_INDEFINITE_WAIT );
  if ( ulRC != NO_ERROR )
    debug( "DosRequestMutexSem(), rc = %u", ulRC );
  lnkseqRemove( &lsCCList, pCC );
  iRFBTid = pCC->iRFBTid;
  DosReleaseMutexSem( hmtxCCList );

  if ( iRFBTid == -1 )
    debug( "No thread for the object" );
  else if ( ccDisconnectSignal( pCC ) )
  {
    ULONG  ulTimerStart, ulTimer;
    BOOL   fFinished = FALSE;

    debugCP( "Wait for thread" );
    DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &ulTimerStart,
                     sizeof(ULONG) );
    do
    {
      ulRC = DosWaitThread( (PULONG)&iRFBTid, DCWW_NOWAIT );
      if ( ulRC == NO_ERROR )
      {
        fFinished = TRUE;
        break;
      }
      if ( ulRC == ERROR_INTERRUPT )
        break;

      DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &ulTimer, sizeof(ULONG) );
    }
    while( ( DosSleep( 1 ) == NO_ERROR ) &&
           ( (ulTimer - ulTimerStart) < 500 ) );
      
    if ( !fFinished )
    {
      debug( "Thread is not finished, kill it..." );
      ulRC = DosKillThread( iRFBTid );
      if ( ulRC == NO_ERROR )
      {
        debugDec( "$threads" );  // The _endthread() function was not called...
        _ccSetThreadState( pCC, RFBSTATE_FINISH );
      }
      else
        debug( "DosKillThread(), rc = %u", ulRC );
    }

    debugCP( "Thread ended, continue destruction..." );
  }

  _RFBClientCleanUp( pCC->pRFBClient );

#ifndef _LASTLOGENTRYONLY
  lnkseqFree( &pCC->lsLog, PLOGREC, free );
#else
  if ( pCC->pcLastLogEnt != NULL )
    free( pCC->pcLastLogEnt );
#endif

  _ccDestoryBitmap( pCC );

  if ( pCC->pLogonInfo != NULL )
    free( pCC->pLogonInfo );

  if ( pCC->hptrPointer != NULLHANDLE )
    WinDestroyPointer( pCC->hptrPointer );

  if ( pCC->hmtxRFB != NULLHANDLE )
    DosCloseMutexSem( pCC->hmtxRFB );

  if ( pCC->pRFBKeySeq != NULL )
    free( pCC->pRFBKeySeq );

  free( pCC );
}

HWND ccSetWindow(PCLNTCONN pCC, HWND hwnd)
{
  HWND       hwndOld = pCC->hwnd;
/*  ULONG      ulRC = DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );

  if ( ulRC != NO_ERROR )
    debug( "DosRequestMutexSem(), rc = %u", ulRC );
*/
  if ( hwnd != NULLHANDLE )
    pCC->hwnd = hwnd;
/*  DosReleaseMutexSem( pCC->hmtxRFB );*/

  return hwndOld;
}

BOOL ccSetViewOnly(PCLNTCONN pCC, ULONG ulViewOnly)
{
  BOOL       fViewOnly;

  switch( ulViewOnly )
  {
    case CCVO_OFF:
      fViewOnly = FALSE;
      break;

    case CCVO_ON:
      fViewOnly = TRUE;
      break;

    default:  // CCVO_TOGGLE:
      pCC->pRFBClient->appData.viewOnly = !pCC->pRFBClient->appData.viewOnly;
      return TRUE;
  }

  if ( !pCC->pRFBClient->appData.viewOnly == !fViewOnly )
    return FALSE;

  pCC->pRFBClient->appData.viewOnly = fViewOnly;

  return TRUE;
}

/*
  BOOL ccIsRFBMsgSupported(PCLNTCONN pCC, BOOL fClient2Server, ULONG ulFRBMsg)

  Returns TRUE if RFB message type ulFRBMsg is supported. Defines for ulFRBMsg
  (see ..\libvncserver\rfb\rfbproto.h):

  fClient2Server is TRUE:
    rfbSetPixelFormat           rfbFixColourMapEntries
    rfbSetEncodings             rfbFramebufferUpdateRequest
    rfbKeyEvent                 rfbPointerEvent
    rfbClientCutText            rfbFileTransfer
    rfbSetScale                 rfbSetServerInput
    rfbSetSW                    rfbTextChat
    rfbPalmVNCSetScaleFactor    rfbXvp

  fClient2Server is FALSE:
    rfbFramebufferUpdate        rfbSetColourMapEntries
    rfbBell                     rfbServerCutText
    rfbResizeFrameBuffer        rfbPalmVNCReSizeFrameBuffer
*/
BOOL ccIsRFBMsgSupported(PCLNTCONN pCC, BOOL fClient2Server, ULONG ulFRBMsg)
{
  return fClient2Server ?
           SupportsClient2Server( pCC->pRFBClient, ulFRBMsg ) :
           SupportsServer2Client( pCC->pRFBClient, ulFRBMsg );
}

// Returns TRUE if current state is RFBSTATE_READY. Ensures that state will
// not change until ccUnlockReadyState().
BOOL ccLockReadyState(PCLNTCONN pCC)
{
  DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );
  if ( pCC->ulRFBState != RFBSTATE_READY )
  {
    DosReleaseMutexSem( pCC->hmtxRFB );
    return FALSE;
  }

  return TRUE;
}

// Unlocks state change from RFBSTATE_READY to another.
VOID ccUnlockReadyState(PCLNTCONN pCC)
{
  DosReleaseMutexSem( pCC->hmtxRFB );
}

/* Sends command to the thread to end session.
   Returns FALSE if a signal has already been sent to the object or the object
   is in the final state.  */
BOOL ccDisconnectSignal(PCLNTCONN pCC)
{
  BOOL  fSendSignal;

  DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );
  fSendSignal = ( pCC->ulStopSigTime == 0 ) && ( pCC->ulRFBState != RFBSTATE_FINISH );
  if ( fSendSignal )
  {
    DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &pCC->ulStopSigTime,
                     sizeof(ULONG) );
    if ( pCC->ulStopSigTime == 0 )  // Avoid zero value.
      pCC->ulStopSigTime++;
    if ( pCC->pRFBClient->sock != RFB_INVALID_SOCKET )
      so_cancel( pCC->pRFBClient->sock );
    debug( "Set disconnect signal at %u msec.", pCC->ulStopSigTime );
  }
  DosReleaseMutexSem( pCC->hmtxRFB );

  return fSendSignal;
}

HPS ccGetHPS(PCLNTCONN pCC)
{
  DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );
  return pCC->hpsMem;
}

VOID ccReleaseHPS(PCLNTCONN pCC, HPS hps)
{
  if ( hps != pCC->hpsMem )
  {
    debug( "Invalid HPS" );
    return;
  }

  DosReleaseMutexSem( pCC->hmtxRFB );
}

BOOL ccQueryFrameSize(PCLNTCONN pCC, PSIZEL psizeFrame)
{
  if ( !ccLockReadyState( pCC ) )
    return FALSE;

  psizeFrame->cx = pCC->pRFBClient->width;
  psizeFrame->cy = pCC->pRFBClient->height;

  ccUnlockReadyState( pCC );
  return TRUE;
}

BOOL ccQueryViewOnly(PCLNTCONN pCC)
{
  return pCC->pRFBClient->appData.viewOnly;
}

BOOL ccQueryTLS(PCLNTCONN pCC)
{
  return pCC->pRFBClient->tlsSession != NULL;
}

HPOINTER ccQueryPointer(PCLNTCONN pCC)
{
  return pCC->pRFBClient->appData.viewOnly || ( pCC->hptrPointer == NULLHANDLE )
           ? WinQuerySysPointer( HWND_DESKTOP, SPTR_ARROW, FALSE )
           : pCC->hptrPointer;
}

// Writes in pcBuf information specified by ulItem from VNC session
BOOL ccQuerySessionInfo(PCLNTCONN pCC, ULONG ulItem, ULONG cbBuf, PCHAR pcBuf)
{
  BOOL  fSuccess = FALSE;
  PSZ   pszText = NULL;

  if ( ulItem > CCSI_SERVER_HOST )
    return FALSE;

  DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );

  switch( ulItem )
  {
    case CCSI_LAST_LOG_REC:
    {
#ifdef _LASTLOGENTRYONLY
      pszText = pCC->pcLastLogEnt;
#else
      PLOGREC    pLogRec = (PLOGREC)lnkseqGetFirst( &pCC->lsLog );

      if ( pLogRec != NULL )
        pszText = pLogRec->acText;
#endif
      break;
    }

    case CCSI_DESKTOP_NAME:
      pszText = pCC->pRFBClient->desktopName;
      break;

    case CCSI_SERVER_HOST:
      pszText = pCC->pRFBClient->serverHost;
      break;
  } // switch

  if ( pszText != NULL )
  {
    strlcpy( pcBuf, pszText, cbBuf );
    fSuccess = TRUE;
  }

  DosReleaseMutexSem( pCC->hmtxRFB );

  if ( !fSuccess && ( cbBuf != 0 ) )
    *pcBuf = '\0';

  return fSuccess;
}

VOID ccSendMouseEvent(PCLNTCONN pCC, LONG lX, LONG lY, ULONG ulButton)
{
  DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );

  if ( !pCC->pRFBClient->appData.viewOnly &&
       ( pCC->ulRFBState == RFBSTATE_READY ) )
  {
    if ( ulButton != 0 )
    {
      int   iRFBButton;

      switch( ulButton & ~RFBBUTTON_PRESSED )
      {
        case RFBBUTTON_LEFT:       iRFBButton = rfbButton1Mask; break;
        case RFBBUTTON_RIGHT:      iRFBButton = rfbButton3Mask; break;
        case RFBBUTTON_MIDDLE:     iRFBButton = rfbButton2Mask; break;
        case RFBBUTTON_WHEEL_DOWN: iRFBButton = rfbButton5Mask; break;
//        case RFBBUTTON_WHEEL_UP:
        default:                   iRFBButton = rfbButton4Mask; break;
      }

      if ( (ulButton & RFBBUTTON_PRESSED) != 0 )
        pCC->iRFBButtons |= iRFBButton;
      else
        pCC->iRFBButtons &= ~iRFBButton;
    }

    lY = pCC->pRFBClient->height - lY - 1;
    SendPointerEvent( pCC->pRFBClient, lX, lY, pCC->iRFBButtons );

    // Only "pressed" events for Wheel.
    if ( (pCC->iRFBButtons & (rfbButton4Mask | rfbButton5Mask)) != 0 )
    {
      pCC->iRFBButtons &= ~(rfbButton4Mask | rfbButton5Mask);
      SendPointerEvent( pCC->pRFBClient, lX, lY, pCC->iRFBButtons );
    }
  }

  DosReleaseMutexSem( pCC->hmtxRFB );
}

VOID ccSendKeyEvent(PCLNTCONN pCC, ULONG ulKey, BOOL fDown)
{
  DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );

  if ( !pCC->pRFBClient->appData.viewOnly &&
       ( pCC->ulRFBState == RFBSTATE_READY ) )
    SendKeyEvent( pCC->pRFBClient, ulKey, fDown );

  DosReleaseMutexSem( pCC->hmtxRFB );
}

VOID ccSendWMCharEvent(PCLNTCONN pCC, MPARAM mp1, MPARAM mp2)
{
  ULONG      ulIdx;

  if ( mp1 == 0
         ? xkKeysymFromMPCheck( &pCC->stXKFromMP )
         : ( xkKeysymFromMP( pKbdMap, mp1, mp2, &pCC->stXKFromMP ) !=
              XKEYMETHOD_NOTFOUND )
     )
  {
    for( ulIdx = 0; ulIdx < pCC->stXKFromMP.cOutput; ulIdx++ )
      ccSendKeyEvent( pCC, pCC->stXKFromMP.aOutput[ulIdx].ulKeysym,
                      pCC->stXKFromMP.aOutput[ulIdx].fPressed );
  }
}

// Sends XK-key codes. For pressed state code must be ORed with CCKEY_PRESS.
VOID ccSendKeySequence(PCLNTCONN pCC, ULONG cKeySeq, PULONG paulKeySeq)
{
  PKEYSEQ    pRFBKeySeq;

  DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );

  if ( !pCC->pRFBClient->appData.viewOnly )
  {
    if ( pCC->pRFBKeySeq == NULL )
      pRFBKeySeq = calloc( 1, sizeof(KEYSEQ) - sizeof(ULONG) +
                           ( cKeySeq * sizeof(ULONG) ) );
    else
    {
      debugPCP( "Previous key sequence was not sent fully" );

      pRFBKeySeq = realloc( pCC->pRFBKeySeq,
                            sizeof(KEYSEQ) - sizeof(ULONG) +
                            ( (pCC->pRFBKeySeq->ulCount + cKeySeq) * sizeof(ULONG) ) );
    }

    if ( pRFBKeySeq != NULL )
    {
      memcpy( &pRFBKeySeq->aulKeys[pRFBKeySeq->ulCount], paulKeySeq,
              cKeySeq * sizeof(ULONG) );
      pRFBKeySeq->ulCount += cKeySeq;
      pCC->pRFBKeySeq = pRFBKeySeq;
    }
  }

  DosReleaseMutexSem( pCC->hmtxRFB );
}

VOID ccSendClipboardText(PCLNTCONN pCC, PSZ pszText)
{
  if ( pszText == NULL )
    return;

  DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );

  if ( !pCC->pRFBClient->appData.viewOnly &&
       ( pCC->ulRFBState == RFBSTATE_READY ) && ( pszText != NULL ) )
  {
    PCHAR  pcSrc, pcDst;
    ULONG  cbText;

    // Remove all CRs
    for( pcSrc = pszText, pcDst = pszText; *pcSrc != '\0'; pcSrc++ )
    {
      if ( *pcSrc == '\r' )
        continue;

      *pcDst = *pcSrc;
      pcDst++;
    }
    cbText = pcDst - pszText;

#if 1
    pcDst = ccLRConvNew( pCC->pRFBClient, CNV_LOCALTOREMOTE, cbText, pszText,
                         &cbText );

    if ( pcDst != NULL )
    {
      debug( "\"%.*s\" -> \"%s\"", cbText, pszText, pcDst );
      SendClientCutText( pCC->pRFBClient, pcDst, cbText );
      free( pcDst );
    }
    else
    {
      debug( "ccLRConvNew(,,\"%.*s\",) failed", cbText, pszText );
      SendClientCutText( pCC->pRFBClient, pszText, cbText );
    }
#else
    SendClientCutTextUTF8( pCC->pRFBClient, pszText, cbText );
#endif
  }

  DosReleaseMutexSem( pCC->hmtxRFB );
}

BOOL ccSendChat(PCLNTCONN pCC, ULONG ulCmd, PSZ pszText)
{
  BOOL       fSuccess = FALSE;

  DosRequestMutexSem( pCC->hmtxRFB, SEM_INDEFINITE_WAIT );

  if ( pCC->ulRFBState == RFBSTATE_READY )
  {
    switch( ulCmd )
    {
      case CCCHAT_OPEN:
        pCC->pRFBClient->supportedMessages.client2server[((rfbTextChat & 0xFF)/8)]
          |= (1<<(rfbTextChat % 8));
        fSuccess = TextChatOpen( pCC->pRFBClient );
        break;

      case CCCHAT_CLOSE:
        fSuccess = TextChatClose( pCC->pRFBClient ) &&
                   TextChatFinish( pCC->pRFBClient );
        break;

      case CCCHAT_MESSAGE:
        {
          PSZ  pszEncText = ccLRConvNew( pCC->pRFBClient, CNV_LOCALTOREMOTE,
                                         -1, pszText, NULL );

          fSuccess = TextChatSend( pCC->pRFBClient,
                                   pszEncText == NULL ? pszText : pszEncText );
          if ( pszEncText != NULL )
            free( pszEncText );
        }
        break;
    }
  }

  DosReleaseMutexSem( pCC->hmtxRFB );

  return fSuccess;
}

BOOL ccFXRequestFileList(PCLNTCONN pCC, PSZ pszPath)
{
  return fxFileListRequest( pCC->pRFBClient, pszPath );
}

// Start file receiving.
BOOL ccFXRecvFile(PCLNTCONN pCC, PSZ pszRemoteName, PSZ pszLocalName)
{
  return fxRecvFile( pCC->pRFBClient, pszRemoteName, pszLocalName );
}

BOOL ccFXAbortFileTransfer(PCLNTCONN pCC)
{
  return fxAbortFileTransfer( pCC->pRFBClient );
}

BOOL ccFXDelete(PCLNTCONN pCC, PSZ pszRemoteName)
{
  return fxDelete( pCC->pRFBClient, pszRemoteName );
}

BOOL ccFXMkDir(PCLNTCONN pCC, PSZ pszRemoteName)
{
  return fxMkDir( pCC->pRFBClient, pszRemoteName );
}

BOOL ccFXRename(PCLNTCONN pCC, PSZ pszRemoteName, PSZ pszNewName)
{
  return fxRename( pCC->pRFBClient, pszRemoteName, pszNewName );
}

BOOL ccFXSendFile(PCLNTCONN pCC, PSZ pszRemoteName, PSZ pszLocalName)
{
  return fxSendFile( pCC->pRFBClient, pszRemoteName, pszLocalName );
}

VOID ccThreadWatchdog()
{
  ULONG      ulTime;
  PCLNTCONN  pCC;
  ULONG      ulRC;

  DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &ulTime, sizeof(ULONG) );

  DosRequestMutexSem( hmtxCCList, SEM_INDEFINITE_WAIT );

  for( pCC = (PCLNTCONN)lnkseqGetFirst( &lsCCList ); pCC != NULL;
       pCC = (PCLNTCONN)lnkseqGetNext( pCC ) )
  {
    if ( ( pCC->ulStopSigTime != 0 ) &&
         ( (ulTime - pCC->ulStopSigTime) > _WAIT_THREAD_TIME ) )
    {
      debug( "The thread does not end for a long time (%u msec.), "
             "kill it...", ulTime - pCC->ulStopSigTime );
      _ccSetThreadState( pCC, RFBSTATE_FINISH );
      ulRC = DosKillThread( pCC->iRFBTid );
      if ( ulRC != NO_ERROR )
        debug( "DosKillThread(), rc = %u", ulRC );
      else
      {
        debugDec( "$threads" );  // The _endthread() function was not called...
        _ccSetThreadState( pCC, RFBSTATE_FINISH );
      }
    }
  }

  DosReleaseMutexSem( hmtxCCList );
}

BOOL ccSearchListener(PSZ pszAddress, USHORT usPort)
{
  PCLNTCONN  pCC;

  DosRequestMutexSem( hmtxCCList, SEM_INDEFINITE_WAIT );

  for( pCC = (PCLNTCONN)lnkseqGetFirst( &lsCCList ); pCC != NULL;
       pCC = (PCLNTCONN)lnkseqGetNext( pCC ) )
  {
    if ( ( pCC->pRFBClient->listenPort == usPort ) &&
         ( STR_ICMP( pCC->pRFBClient->listenAddress, pszAddress ) == 0 ) )
    {
      if ( pCC->hwnd != NULLHANDLE )
      {
        WinSetWindowPos( pCC->hwnd, HWND_TOP, 0, 0, 0, 0,
                         SWP_ZORDER | SWP_ACTIVATE );
        break;
      }
    }
  }

  DosReleaseMutexSem( hmtxCCList );

  return pCC != NULL;
}
