#include <stdlib.h>
#include <types.h>
#include <stdio.h>
#define INCL_DOSERRORS
#define INCL_WIN
#define INCL_WINDIALOGS
#define INCL_GPI
#include <os2.h>
#include <utils.h>
#include <rfb/rfbclient.h>
#include "clntconn.h"
#include "clntwnd.h"
#include "lnchpad.h"
#include "pswddlg.h"
#include "resource.h"
#include "prbar.h"
#include "progress.h"
#include <debug.h>

#define TIMER_POPUP_ID           1
// TIMER_POPUP_TIMEOUT: Delay before appearance of the progress dialog [msec].
#define TIMER_POPUP_TIMEOUT      1500
#define TIMER_RECONNECT_ID       2
// TIMER_RECONNECT_TIMEOUT: Pause before next connection attempt [msec].
#define TIMER_RECONNECT_TIMEOUT  3000

#define WM_PR_SWITCH_TO          (WM_USER + 1)

typedef struct _DLGINITDATA {
  USHORT               usSize;
  PHOSTDATA            pHostData;
} DLGINITDATA, *PDLGINITDATA;

typedef struct _DLGDATA {
  HOSTDATA             stHostData;
  ULONG                ulAttempt;
  PCLNTCONN            pCC;
  PSZ                  pszPrBarText;
  ULONG                ulTimerPopupId;
  ULONG                ulTimerReconnectId;
  BOOL                 fUseMemorizedPswd;
  PCCLOGONINFO         pLogonInfo;
} DLGDATA, *PDLGDATA;


extern HAB             hab;
extern HMQ             hmq;
extern ULONG           cOpenWin;

static DLGINITDATA     stInitData;


// Progress dialog
// ---------------

// Creates a new RFB-protocol connection object.
static BOOL _startConnection(HWND hwnd)
{
  PDLGDATA   pDlgData = (PDLGDATA)WinQueryWindowPtr( hwnd, QWL_USER );
  HWND       hwndBar = WinWindowFromID( hwnd, IDPBAR_PROGRESS );
  CHAR       acBuf[128];

  if ( pDlgData->pCC != NULL )
  {
    debugCP( "ccDestroy()..." );
    ccDestroy( pDlgData->pCC );
  }

  debugCP( "ccCreate()..." );
  pDlgData->pCC = ccCreate( &pDlgData->stHostData.stProperties, hwnd );
  if ( pDlgData->pCC == NULL )
  {
    debugCP( "ccCreate() failed" );
    WinMessageBox( HWND_DESKTOP, hwnd, "Internal programm error", APP_NAME, 0,
                   MB_ERROR | MB_OK );
    return FALSE;
  }

  pDlgData->ulAttempt++;
  debug( "pDlgData->ulAttempt=%lu", pDlgData->ulAttempt );

  if ( WinSubstituteStrings( hwnd, pDlgData->pszPrBarText, sizeof(acBuf),
                             acBuf ) != 0 )
    // Substitution process on PogressBar text for key %2 (attempts).
    WinSetWindowText( hwndBar, acBuf );

  return TRUE;
}

static VOID _showProgress(HWND hwnd, PDLGDATA pDlgData)
{
  POINTL  pt;
  ULONG   ulFlags;

  if ( !WinIsWindowVisible( hwnd ) )
  {
    // Move the window to the center of the desktop only if it was invisible.
    RECTL   rectDT, rectWin;

    WinQueryWindowRect( hwnd, &rectWin );
    WinQueryWindowRect( HWND_DESKTOP, &rectDT );
    pt.x = ( rectDT.xRight - ( rectWin.xRight - rectWin.xLeft ) ) / 2;
    pt.y = ( rectDT.yTop -   ( rectWin.yTop - rectWin.yBottom ) ) / 2;
    ulFlags = SWP_ACTIVATE | SWP_ZORDER | SWP_SHOW | SWP_MOVE;
  }
  else
    /* The window has already been displayed, we bring it to the foreground
       without moving it to the center.  */
    ulFlags = SWP_ACTIVATE | SWP_ZORDER;

  WinSetWindowPos( hwnd, HWND_TOP, pt.x, pt.y, 0, 0, ulFlags );
}

static BOOL _wmInitDlg(HWND hwnd, PDLGINITDATA pInitData)
{
  PDLGDATA        pDlgData = calloc( 1, sizeof(DLGDATA) );
  HWND            hwndBar = WinWindowFromID( hwnd, IDPBAR_PROGRESS );
  POINTL          pt;
  RECTL           rectDT, rectWin;
  CHAR            acBuf[256];
  CHAR            acTitle[256];
  PRBARINFO       stPrBarInfo;

  if ( pDlgData == NULL )
  {
    WinDestroyWindow( hwnd );
    return FALSE;
  }

  lpQueryWinPresParam( hwnd );

  WinSetWindowPtr( hwnd, QWL_USER, pDlgData );
  pDlgData->stHostData        = *pInitData->pHostData;
  pDlgData->ulTimerPopupId    = 0;
  pDlgData->fUseMemorizedPswd = pInitData->pHostData->fRememberPswd;

  // Set window title.
  WinQueryWindowText( hwnd, sizeof(acBuf), acBuf );
  WinSubstituteStrings( hwnd, acBuf, sizeof(acTitle), acTitle ); // %0=hostname
  WinSetWindowText( hwnd, acTitle );

  // Place the window in the center of the desktop.
  WinQueryWindowRect( hwnd, &rectWin );
  WinQueryWindowRect( HWND_DESKTOP, &rectDT );
  pt.x = ( rectDT.xRight - ( rectWin.xRight - rectWin.xLeft ) ) / 2;
  pt.y = ( rectDT.yTop -   ( rectWin.yTop - rectWin.yBottom ) ) / 2;
  WinSetWindowPos( hwnd, HWND_TOP, pt.x, pt.y, 0, 0,
                   SWP_ACTIVATE | SWP_ZORDER | SWP_HIDE | SWP_MOVE );

  // Store ProgressBar text with key %2. It is not replaced now (see function
  // _wmSubstituteString()) but will be at each connection (_startConnection()).
  if ( WinQueryWindowText( hwndBar, sizeof(acBuf), acBuf ) != 0 )
    pDlgData->pszPrBarText = strdup( acBuf );

  if ( !_startConnection( hwnd ) )
  {
    debugCP( "_startConnection() failed" );
    WinDestroyWindow( hwnd );
    return FALSE;
  }

  if ( pDlgData->stHostData.stProperties.lListenPort >= 0 )
  {
    // Progress bar text in "listen mode".

    WinLoadString( hab, 0, IDS_WAIT_SERVER, sizeof(acBuf), acBuf );
    WinSetWindowText( hwndBar, acBuf );
    stPrBarInfo.ulAnimation = PRBARA_RIGHT;

    // Popup immediately
    _showProgress( hwnd, pDlgData );
  }
  else
  {
    pDlgData->ulTimerPopupId = WinStartTimer( hab, hwnd, TIMER_POPUP_ID,
                                              TIMER_POPUP_TIMEOUT );
    stPrBarInfo.ulAnimation = PRBARA_LEFTDBL;
  }

  WinSendMsg( hwndBar, PBM_SETPARAM, MPFROMLONG(PBARSF_ANIMATION),
              MPFROMP(&stPrBarInfo) );

  return TRUE;
}

static VOID _wmDestroy(HWND hwnd)
{
  PDLGDATA        pDlgData = (PDLGDATA)WinQueryWindowPtr( hwnd, QWL_USER );

  lpStoreWinPresParam( hwnd );

  if ( pDlgData == NULL )
    return;

  if ( pDlgData->ulTimerPopupId != 0 )
    WinStopTimer( hab, hwnd, pDlgData->ulTimerPopupId );

  if ( pDlgData->ulTimerReconnectId != 0 )
    WinStopTimer( hab, hwnd, pDlgData->ulTimerReconnectId );

  if ( pDlgData->pszPrBarText != NULL )
    free( pDlgData->pszPrBarText );

  free( pDlgData );
}

static PSZ _wmSubstituteString(HWND hwnd, USHORT usKey)
{
  PDLGDATA             pDlgData = (PDLGDATA)WinQueryWindowPtr( hwnd, QWL_USER );
  static CHAR          acBuf[32];

  switch( usKey )
  {
    case 0: // %0 - host name.
      if ( pDlgData != NULL )
      {
        // The function called during WM_INITDLG for the window title.

        if ( pDlgData->stHostData.stProperties.lListenPort < 0 )
          return pDlgData->stHostData.stProperties.acHost;

        // "Listening mode"
        sprintf( acBuf, "%s:%u",
                 pDlgData->stHostData.stProperties.acHost[0] == '\0'
                   ? "*" : pDlgData->stHostData.stProperties.acHost,
                 pDlgData->stHostData.stProperties.lListenPort );
        return acBuf;
      }

    case 1: // %1 - maximum connect attempts (ProgressBar text).
      return ultoa( stInitData.pHostData->ulAttempts, acBuf, 10 ); 

    case 2: // %2 - number of connect attempt (ProgressBar text).
      // We don't touch key %2 in dialog loading time. Only when dialog data
      // installed, from _startConnection().
      if ( pDlgData != NULL )
        return ultoa( pDlgData->ulAttempt, acBuf, 10 ); 
  }

  return NULL;
}

static BOOL _wmTimer(HWND hwnd, ULONG ulTimerId)
{
  PDLGDATA   pDlgData = (PDLGDATA)WinQueryWindowPtr( hwnd, QWL_USER );

  if ( ulTimerId == TIMER_POPUP_ID ) //pDlgData->ulTimerPopupId )
  {
    // Stop one-event timer for show (popup) window after creation.
    WinStopTimer( hab, hwnd, pDlgData->ulTimerPopupId );
    pDlgData->ulTimerPopupId = 0;

    _showProgress( hwnd, pDlgData );
  }
  else if ( ulTimerId == TIMER_RECONNECT_ID ) //pDlgData->ulTimerReconnectId )
  {
    // Stop one-event timer for reconnect time-out.
    PRBARINFO          stPrBarInfo;

    WinStopTimer( hab, hwnd, TIMER_RECONNECT_ID ); //pDlgData->ulTimerReconnectId );
    pDlgData->ulTimerReconnectId = 0;

    stPrBarInfo.ulImageIdx = 0;
    WinSendDlgItemMsg( hwnd, IDPBAR_PROGRESS, PBM_SETPARAM,
                       MPFROMLONG(PBARSF_IMAGE), MPFROMP(&stPrBarInfo) );

    WinSetWindowPos( hwnd, HWND_TOP, 0, 0, 0, 0, SWP_ZORDER );
    if ( !_startConnection( hwnd ) )
    {
      debugCP( "_startConnection() failed" );
      WinDestroyWindow( hwnd );
    }
  }
  else
    return FALSE; // Not our timer.

  return TRUE;
}

static VOID _cmdCancel(HWND hwnd)
{
  PDLGDATA  pDlgData = (PDLGDATA)WinQueryWindowPtr( hwnd, QWL_USER );

  if ( pDlgData->pCC != NULL )
  {
    // Hide the window while we (possibly) have a small delay in ccDestroy.
    WinShowWindow( hwnd, FALSE );

    debugCP( "ccDestroy()..." );
    ccDestroy( pDlgData->pCC );
    pDlgData->pCC = NULL;
  }

  WinDestroyWindow( hwnd );
}

static VOID _wmVNCState(HWND hwnd, ULONG ulState)
{
  PDLGDATA   pDlgData = (PDLGDATA)WinQueryWindowPtr( hwnd, QWL_USER );
  PRBARINFO  stPrBarInfo;
  CHAR       acBuf[256];

  switch( ulState )
  {
    case RFBSTATE_INITPROGRESS:
      // After sending the credentials.
      stPrBarInfo.ulAnimation = PRBARA_LEFTDBL;
      WinSendDlgItemMsg( hwnd, IDPBAR_PROGRESS, PBM_SETPARAM,
                         MPFROMLONG(PBARSF_ANIMATION), MPFROMP(&stPrBarInfo) );
      break;

    case RFBSTATE_READY:
    {
      BOOL   fReady = ccLockReadyState( pDlgData->pCC );

      debugCP( "RFBSTATE_READY: create main window, destroy progress dialog" );

      if ( // Logon information typed by user (obtained from dialog).
           ( pDlgData->pLogonInfo != NULL ) &&
           // Need to remember password.
           pDlgData->stHostData.fRememberPswd )
      {
        debugCP( "lpStoreLogonInfo()..." );
        lpStoreLogonInfo( pDlgData->stHostData.stProperties.acHost,
                          pDlgData->pLogonInfo );
        memset( pDlgData->pLogonInfo, 0, sizeof(CCLOGONINFO) );
        free( pDlgData->pLogonInfo );
        pDlgData->pLogonInfo = NULL;
      }

      if ( fReady )
        cwCreate( pDlgData->pCC, pDlgData->stHostData.stProperties.acHost,
                  pDlgData->stHostData.acWinTitle,
                  pDlgData->stHostData.hwndNotify );
      ccUnlockReadyState( pDlgData->pCC );

      WinDestroyWindow( hwnd );
      break;
    }

    case RFBSTATE_WAITUSER:
      // Before requesting credentials.
      stPrBarInfo.ulAnimation = PRBARA_LEFT;
      WinSendDlgItemMsg( hwnd, IDPBAR_PROGRESS, PBM_SETPARAM,
                         MPFROMLONG(PBARSF_ANIMATION), MPFROMP(&stPrBarInfo) );
      break;

    case RFBSTATE_FINISH:
    {
      static PSZ     apszKnownErr[] =
      {
        /* 0 */ "Unable to connect to VNC server",
        /* 1 */ "VNC authentication failed",
        /* 2 */ "VNC connection failed: password check failed!",
        /* 3 */ "read (54: Connection reset by peer)",
        /* 4 */ "VNC connection failed: ",
        /* 5 */ "Reading password failed"
      };
      ULONG  ulIdx;
      BOOL   fLogRec = ccQuerySessionInfo( pDlgData->pCC,
                                  CCSI_LAST_LOG_REC, sizeof(acBuf), acBuf );

      if ( pDlgData->ulTimerPopupId != 0 )
      {
        WinStopTimer( hab, hwnd, TIMER_POPUP_ID );
        pDlgData->ulTimerPopupId = 0;
      }

      if ( pDlgData->ulTimerReconnectId != 0 )
      {
        WinStopTimer( hab, hwnd, TIMER_RECONNECT_ID );
        pDlgData->ulTimerReconnectId = 0;
      }

      if ( fLogRec && ( pDlgData->stHostData.stProperties.lListenPort < 0 )
           && ( pDlgData->ulAttempt < pDlgData->stHostData.ulAttempts ) )
      {
        for( ulIdx = 0; ulIdx < ARRAYSIZE(apszKnownErr); ulIdx++ )
        {
          if ( memcmp( acBuf, apszKnownErr[ulIdx],
                       strlen( apszKnownErr[ulIdx] ) ) == 0 )
            break; // end for()
        }

        if ( ulIdx < ARRAYSIZE(apszKnownErr) )
        {
          BOOL  fRetry;

          /* We don't use stored password for next connection attempts
             (will ask user) if error other than
             "Unable to connect to VNC server".  */
          pDlgData->fUseMemorizedPswd &= ( ulIdx == 0 );

          if ( pDlgData->pLogonInfo != NULL )
          {
            memset( pDlgData->pLogonInfo, 0, sizeof(CCLOGONINFO) );
            free( pDlgData->pLogonInfo );
            pDlgData->pLogonInfo = NULL;
          }

          if ( ( ulIdx == 1 ) || ( ulIdx == 2 ) )
          {
            /* Error "VNC authentication failed" - inform the user that
               authentication failed and prompt to try again.  */
            fRetry = utilMessageBox( hwnd, APP_NAME, IDM_AUTHFAILED,
                              MB_ICONHAND | MB_MOVEABLE | MB_RETRYCANCEL ) ==
                       MBID_RETRY;
          }
          else if ( ulIdx == 4 )
          {
            /* Error "VNC connection failed: <reason>" - show the text to
               the user and prompt to try again.  */
            fRetry = WinMessageBox( HWND_DESKTOP, hwnd, acBuf, APP_NAME, 0,
                           MB_ERROR | MB_MOVEABLE | MB_RETRYCANCEL ) ==
                       MBID_RETRY;
          }
          else
            fRetry = ulIdx != 5;
            /* 5 - "Reading password failed"
               When the user's password/credentials are canceled in
               _wmVNCQueryLogon(), NULL is returned to vncserverclient (see
               _cbRFBGetPassword() and _cbRFBGetCredential() in clntconn.c).
               In this case, the message "Reading password failed" is written
               to the log by vncserverclient. There is no point in showing this
               to the user or reconnect.  */

          if ( fRetry )
          {
            stPrBarInfo.ulImageIdx = 1;
            WinSendDlgItemMsg( hwnd, IDPBAR_PROGRESS, PBM_SETPARAM,
                              MPFROMLONG(PBARSF_IMAGE), MPFROMP(&stPrBarInfo) );
            _showProgress( hwnd, pDlgData );

            // Start reconnect timer.
            pDlgData->ulTimerReconnectId =
                                  WinStartTimer( hab, hwnd, TIMER_RECONNECT_ID,
                                                 TIMER_RECONNECT_TIMEOUT );
            break; // end case RFBSTATE_FINISH:
          }
          fLogRec = FALSE;
        }
      }

      if ( fLogRec )
      {
        /* There is some error, after which we can not continue to try to
           connect or it was last attempt. We show last log message to the
           user.  */
        WinMessageBox( HWND_DESKTOP, hwnd, acBuf, APP_NAME, 0,
                       MB_ERROR | MB_MOVEABLE | MB_OK );
      }

      debugCP( "ccDestroy()..." );
      ccDestroy( pDlgData->pCC );
      pDlgData->pCC = NULL;
      WinDestroyWindow( hwnd );
    }
    break;
  }  // switch( ulState )
}

static BOOL _wmVNCQueryLogon(HWND hwnd, PCCLOGONINFO pLogonInfo)
{
  PDLGDATA  pDlgData = (PDLGDATA)WinQueryWindowPtr( hwnd, QWL_USER );

  debugCP();

  if ( pDlgData->fUseMemorizedPswd )
  {
    if ( lpQueryLogonInfo( pDlgData->stHostData.stProperties.acHost,
                           pLogonInfo, pLogonInfo->fCredential ) )
    {
      debug( "Send stored logon information (%s)...",
             pLogonInfo->fCredential ? "credential" : "password" );
      return TRUE;
    }

    debug( "srvdlgQueryHostPswd() failed" );
    pDlgData->fUseMemorizedPswd = FALSE;
  }

  _wmTimer( hwnd, pDlgData->ulTimerPopupId );
  WinSetWindowPos( hwnd, HWND_TOP, 0, 0, 0, 0, SWP_ZORDER );

  if ( !pswdlgRequest( pDlgData->pCC, hwnd, pLogonInfo ) )
  {
    debugCP( "Password dialog has been canceled" );
    return FALSE;
  }

  debug( "%s received from password dialog, send it to the connection object...",
         pLogonInfo->fCredential ? "Credential" : "Password" );

  if ( pDlgData->stHostData.fRememberPswd )
  {
    if ( pDlgData->pLogonInfo )
      free( pDlgData->pLogonInfo );
    pDlgData->pLogonInfo = malloc( sizeof(CCLOGONINFO) );

    if ( pDlgData->pLogonInfo != NULL )
      memcpy( pDlgData->pLogonInfo, pLogonInfo, sizeof(CCLOGONINFO) );
  }

  return TRUE;
}

static MRESULT EXPENTRY _dlgProc(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
  switch( msg )
  {
    case WM_INITDLG:
      cOpenWin++;
      debug( "WM_CREATE, cOpenWin=%lu", cOpenWin );
      return (MRESULT)_wmInitDlg( hwnd, (PDLGINITDATA)mp2 );

    case WM_DESTROY:
      _wmDestroy( hwnd );
      cOpenWin--;
      debug( "WM_DESTROY, cOpenWin=%lu", cOpenWin );
      if ( cOpenWin == 0 )
        WinPostQueueMsg( hmq, WM_QUIT, 0, 0 );
      break;

    case WM_SUBSTITUTESTRING:
      return (MRESULT)_wmSubstituteString( hwnd, SHORT1FROMMP(mp1) );

    case WM_TIMER:
      if ( _wmTimer( hwnd, SHORT1FROMMP(mp1) ) )
        return (MRESULT)TRUE;
      break;

    case WM_COMMAND:
      switch( SHORT1FROMMP( mp1 ) )
      {
        case MBID_CANCEL:
          _cmdCancel( hwnd );
          break;
      }
      return (MRESULT)0;

    case WM_VNC_STATE:
      _wmVNCState( hwnd, LONGFROMMP(mp2) );
      return (MRESULT)TRUE;

    case WM_VNC_QUERYLOGON:
      return (MRESULT)_wmVNCQueryLogon( hwnd, (PCCLOGONINFO)mp2 );
  }

  return WinDefDlgProc( hwnd, msg, mp1, mp2 );
}


// Should be called from the main thread only.
BOOL prStart(PHOSTDATA pHostData)
{
  HWND         hwndDlg;

  stInitData.usSize     = sizeof(DLGINITDATA);
  stInitData.pHostData  = pHostData;

  hwndDlg = WinLoadDlg( HWND_DESKTOP, HWND_DESKTOP, _dlgProc, NULLHANDLE,
                        IDDLG_PROGRESS, &stInitData );
  if ( hwndDlg == NULLHANDLE )
  {
    debug( "WinLoadDlg() failed" );
    return FALSE;
  }

  return TRUE;
}
