#define INCL_DOSERRORS
#define INCL_WIN
#define INCL_WINDIALOGS
#include <os2.h>
#include <types.h>
#include "resource.h"
#include "clntconn.h"
#include "lnchpad.h"
#include "pswddlg.h"
#include <debug.h>


typedef struct _PSWDDLGDATA {
  USHORT        cb;            /*  Length of control data in bytes. */
  BOOL          fTLS;
  PCCLOGONINFO  pLogonInfo;
} PSWDDLGDATA, *PPSWDDLGDATA;


static BOOL _wmInitDlg(HWND hwnd, PPSWDDLGDATA pData)
{
  HWND       hwndOwner = WinQueryWindow( hwnd, QW_OWNER );
  HWND       hwndCtl;
  POINTL     pointl = { 0 };
  RECTL      rectl;

  lpQueryWinPresParam( hwnd );
  WinSetWindowULong( hwnd, QWL_USER, (ULONG)pData->pLogonInfo );

  WinQueryWindowRect( hwndOwner, &rectl );
  pointl.x = (rectl.xRight - rectl.xLeft) / 4;
  pointl.y = -(rectl.yTop - rectl.yBottom) / 2;
  WinMapWindowPoints( hwndOwner, HWND_DESKTOP, (PPOINTL)&pointl, 1 );

  WinSetWindowPos( hwnd, HWND_TOP, pointl.x, pointl.y, 0, 0,
                   SWP_ACTIVATE | SWP_ZORDER | SWP_SHOW | SWP_MOVE );

  hwndCtl = WinWindowFromID( hwnd, IDEF_USERNAME );
  WinSetFocus( HWND_DESKTOP,
               hwndCtl != NULLHANDLE
                 ? hwndCtl : WinWindowFromID( hwnd, IDEF_PASSWORD ) );

  WinShowWindow( WinWindowFromID( hwnd, IDST_UNENCCONN ), !pData->fTLS );
  WinShowWindow( WinWindowFromID( hwnd, IDST_ENCCONN ), pData->fTLS );

  /* Button "Ok" will be enabled when some characters will be entered in
     username entry field.  */
  WinEnableWindow( WinWindowFromID( hwnd, MBID_OK ), hwndCtl == NULLHANDLE );

  return TRUE;
}

static VOID _cmdOk(HWND hwnd)
{
  PCCLOGONINFO  pLogonInfo =
                  (PCCLOGONINFO)WinQueryWindowULong( hwnd, QWL_USER );
  HWND          hwndCtl = WinWindowFromID( hwnd, IDEF_USERNAME );

  
  if ( hwndCtl != NULLHANDLE )
  {
    pLogonInfo->fCredential = TRUE;
    WinQueryWindowText( hwndCtl, sizeof(pLogonInfo->acUserName),
                        pLogonInfo->acUserName );
    debugCP( "Credential entered" );
  }
  else
  {
    pLogonInfo->fCredential = FALSE;
    pLogonInfo->acUserName[0] = '\0';
    debugCP( "Password entered" );
  }

  pLogonInfo->acPassword[0] = '\0';
  WinQueryDlgItemText( hwnd, IDEF_PASSWORD,
                       sizeof(pLogonInfo->acPassword), pLogonInfo->acPassword );

  WinDismissDlg( hwnd, TRUE );
}

static VOID _cmdCancel(HWND hwnd)
{
  debugCP( "Password enter canceled" );
  WinDismissDlg( hwnd, FALSE );
}

static MRESULT EXPENTRY _dlgProc(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
  switch( msg )
  {
    case WM_INITDLG:
    {
      PPSWDDLGDATA  pDlgData = (PPSWDDLGDATA)mp2;

      return (MRESULT)_wmInitDlg( hwnd, pDlgData );
    }

    case WM_DESTROY:
      lpStoreWinPresParam( hwnd );
      break;

    case WM_CONTROL:
      if ( ( SHORT2FROMMP(mp1) == EN_CHANGE ) &&
           ( WinQueryWindowUShort( HWNDFROMMP(mp2), QWS_ID ) ==
               IDEF_USERNAME ) )
      {
        /* The OK button is activated when there are some characters in the
           username input field.  */
        WinEnableWindow( WinWindowFromID( hwnd, MBID_OK ),
                         WinQueryWindowTextLength( HWNDFROMMP(mp2) ) != 0 );
        break;
      }
      return (MRESULT)TRUE;

    case WM_COMMAND:
      switch( SHORT1FROMMP( mp1 ) )
      {
        case MBID_OK:
          _cmdOk( hwnd );
          return (MRESULT)0;

        case MBID_CANCEL:
          _cmdCancel( hwnd );
          return (MRESULT)0;
      }
      return (MRESULT)TRUE;
  }

  return WinDefDlgProc( hwnd, msg, mp1, mp2 );
}


BOOL pswdlgRequest(PCLNTCONN pCC, HWND hwndOwner, PCCLOGONINFO pLogonInfo)
{
  PSWDDLGDATA  stDlgData;
  BOOL         fRes;

  stDlgData.cb = sizeof(PSWDDLGDATA);
  stDlgData.fTLS = ccQueryTLS( pCC );
  stDlgData.pLogonInfo = pLogonInfo;

  debugCP( "WinDlgBox()..." );
  fRes = (BOOL)WinDlgBox( HWND_DESKTOP, hwndOwner, _dlgProc, NULLHANDLE,
                          pLogonInfo->fCredential
                            ? IDDDLG_CREDENTIAL : IDDLG_PASSWORD,
                          &stDlgData );

  debug( "Result: %s", fRes ? "TRUE" : "FALSE" );
  return fRes;
}

