@echo off

rem List of identifiers for supported languages.
set LangList=en ru

rem Important! Resource compiler version 4 is used.
rem C:\OS2\RC.EXE
rem Version 4.00.011 Oct 10 2000
rem
rem Resource Compiler version 5 is not supported.
rem %unixroot%\usr\bin\rc.exe
rem IBM RC (Resource Compiler) Version 5.00.007 Jan 30 2003


rem --------------------------------------------------


set WPIFile=vnc.wpi

if .%1.==.all. goto all
if .%1.==.src. goto src
if .%1.==.bin. goto bin
if .%1.==.clean. goto clean
if .%1.==.wpi. goto wpi
if .%1.==.newlang. goto newlang

rem For internal calls
if .%1.==.check-env. goto check_env
if .%1.==.clean-ld. goto clean_ld
if .%1.==.get-date. goto get_date
if .%1.==.pack-src. goto pack_src
if .%1.==.make-li. goto make_li
if .%1.==.make-ld. goto make_ld

echo Usage:
echo.
echo %0 all
echo   Cleaning, archive the sources, compile and create WPI in ZIP for each
echo   supported language.
echo.
echo %0 src
echo   Cleaning and archive the sources.
echo.
echo %0 bin ^<%LangList%^>
echo   Make all binaries for the specified language.
echo.
echo %0 wpi ^<%LangList%^>
echo   Make language-depended binaries for the specified language, build WarpIn
echo   package and ZIP it.
echo.
echo %0 newlang
echo   Creating missing language resource files in .\res. Missing *_??.* files will
echo   be copied from the corresponding *_en.* files.
echo.
echo %0 clean
echo   Deleting all files created during compilation.
EXIT


:all
rem --------------------------------------------------
rem Cleaning, archive the sources, compile and create
rem WPI in ZIP for each language.
rem --------------------------------------------------

call %0 src
call %0 make-li

for %%i in (%LangList%) do call %0 wpi %%i 

goto _return


:src
rem  --------------------------------------------------
rem            Cleaning, archive the sources
rem  --------------------------------------------------

call %0 clean
call %0 pack-src
goto _return


:bin
rem  --------------------------------------------------
rem     Make all binaries for the specified language
rem  --------------------------------------------------

if .%2.==.. goto _err_lang_arg

for %%i in (%LangList%) do if %%i==%2 goto _bin_00
echo Unknown target language specified.
echo List of valid target languages: %LangList%.
goto _error
:_bin_00

call %0 check-env
call %0 clean-ld
call %0 make-li
call %0 make-ld %2
goto _return


:wpi
rem --------------------------------------------------
rem                Make WarpIN package
rem --------------------------------------------------

if .%2.==.. goto _err_lang_arg

call %0 clean-ld
call %0 make-ld %2
call %0 get-date

set ARFile=vnc-wpi%ardate%-%2

cd .\WarpIn
call make.cmd %WPIFile%
if not exist %WPIFile% goto _error
cd..

rem Pack WPI
rm -f %ARFile%.zip
7z a -bd -tzip -mx7 -r0 -x!*.zip %ARFile% ./WarpIn/%WPIFile% ./bin/README
if errorlevel 1 goto _error
goto _return


:clean
rem  --------------------------------------------------
rem                       Cleaning
rem  --------------------------------------------------

call %0 check-env

rem Cleaning source directories
set components=libvncserver libgiconv os2xkey KbdXKey VNCServer VNCViewer vnckbd WPS
for %%i in (%components%) do cd %%i & cmd /c make clean & cd ..

rm -f .\res\vncserver\programs.txt

rem Remove files from ./bin
cd .\bin
rm -f *.dbg *.log *.!!! os2xkey.dll vncserver.exe vnckbd.sys
rm -f vncviewer.exe kbdxkey.exe VNCSERVER.INI VNCVIEWER.INI
cd..

rem Remove warpin package
rm -f ./warpin/%WPIFile%
goto _return


:newlang
rem  --------------------------------------------------
rem  Creating missing language resource files in .\res.
rem  --------------------------------------------------

set LDRC=chatwin kbdxkey vncserver vncv vncviewer

cd .\res

for %%i in (%LangList%) do (
  for %%j in (%LDRC%) do (
    if not exist %%j_%%i.rc copy %%j_en.rc %%j_%%i.rc
  )

rem  if not exist .\README_%%i (
rem    copy .\README_en .\README_%%i
rem  )

  if not exist .\vncserver\programs_%%i.txt (
    copy .\vncserver\programs_en.txt .\vncserver\programs_%%i.txt
  )
)

cd ..
goto _return


:check_env
rem  --------------------------------------------------
rem  Checking the environment - the presence of all
rem  necessary utilities, variables, etc.
rem  --------------------------------------------------

if .%unixroot%.==.. (
echo Environment variable UNIXROOT is not set.
goto _error
)

if not exist %unixroot%\usr\libexec\bin\date.exe (
echo %unixroot%\usr\libexec\bin\date.exe utility not found.
goto _error
)

cmd /C rm.exe --help >nul
if errorlevel 1041 (
echo 'rm' utility not found.
goto _error
)

cmd /C make.exe --help >nul
if errorlevel 1041 (
echo 'make' utility not found.
goto _error
)

cmd /C rc.exe >nul
if errorlevel 1041 (
echo Resource compiler 'rc' not found.
goto _error
)

cmd /C lxlite.exe /H >nul
if errorlevel 1041 (
echo lxlite.exe utility not found.
goto _error
)

cmd /C highmem.exe -? >nul
if errorlevel 1041 (
echo highmem.exe utility not found.
goto _error
)

if .%watcom%.==.. (
echo Environment variable WATCOM is not set. Do you have Open Watcom compiler installed?
goto _error
)

cmd /C 7z.exe >nul
if errorlevel 1041 (
echo 7z archiver not found.
goto _error
)

goto _return


:clean_ld
rem  --------------------------------------------------
rem           Remove language dependent files
rem  --------------------------------------------------

rm -f .\bin\KbdXKey.exe .\bin\VNCServer.exe .\bin\VNCViewer.exe .\res\*.res
rm -f .\res\vncserver\programs.txt
goto _return


:get_date
rem  --------------------------------------------------
rem               Set ardate env. variable
rem  --------------------------------------------------

set DateUtility=%unixroot%\usr\libexec\bin\date.exe
rem Query current date (variable %ardate%).
%DateUtility% +"set ardate=-%%Y%%m%%d" >archdate.cmd
call archdate.cmd
rm -f archdate.cmd
goto _return


:pack_src
rem  --------------------------------------------------
rem                 Archive the sources
rem  --------------------------------------------------

call %0 get-date
set ARFile=vnc-src%ardate%

rm -f %ARFile%.zip
7z a -bd -tzip -mx7 -r0 -x!*.zip %ARFile%
goto _return


:make_li
rem  --------------------------------------------------
rem          Make language independent binaries
rem  --------------------------------------------------

set components=libvncserver libgiconv os2xkey
for %%i in (%components%) do @echo Make: %%i & make -s -C ./%%i & if errorlevel 1 goto _error

echo Make: driver vnckbd.sys
cd .\vnckbd
wmake -h
if errorlevel 1 goto _error
cd..
goto _return


:make_ld
rem --------------------------------------------------
rem          Make language dependent binaries
rem --------------------------------------------------

if .%2.==.. goto _err_lang_arg

cmd /C .\res\vncserver\programs.cmd TAGLANG=%2
if errorlevel 1 goto _error

set components=KbdXKey VNCServer VNCViewer
for %%i in (%components%) do @echo Make: %%i & make -s -C ./%%i TAGLANG=%2 & if errorlevel 1 goto _error

echo Make: WPS Class
cd .\WPS
wmake -h TAGLANG=%2 all
if errorlevel 1 goto _error
cd..

rem copy .\res\README_%2 .\bin\README

goto _return


rem --------------------------------------------------

:_err_lang_arg
echo The target language is not specified in the second argument.

:_error
exit 1

:_return